# sage_setup: distribution = sagemath-categories
r"""
Subschemes of projective space

AUTHORS:

- David Kohel (2005): initial version.
- William Stein (2005): initial version.
- Volker Braun (2010-12-24): documentation of schemes and
  refactoring. Added coordinate neighborhoods and is_smooth()
- Ben Hutz (2013) refactoring
"""

# ****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#       Copyright (C) 2013 Ben Hutz <bn4941@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import binomial
from sage.categories.fields import Fields
from sage.categories.homset import Hom
from sage.misc.lazy_import import lazy_import
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import RationalField
from sage.schemes.generic.algebraic_scheme import AlgebraicScheme_subscheme
from sage.schemes.projective.projective_morphism import SchemeMorphism_polynomial_projective_subscheme_field

lazy_import('sage.dynamics.arithmetic_dynamics.generic_ds', 'DynamicalSystem')
lazy_import('sage.matrix.constructor', 'matrix')
lazy_import('sage.schemes.elliptic_curves.ell_generic', 'EllipticCurve_generic', as_='EllipticCurve')


class AlgebraicScheme_subscheme_projective(AlgebraicScheme_subscheme):
    r"""
    Construct an algebraic subscheme of projective space.

    .. WARNING::

        You should not create objects of this class directly. The
        preferred method to construct such subschemes is to use
        :meth:`~sage.schemes.projective.projective_space.ProjectiveSpace_field.subscheme`
        method of :class:`projective space
        <sage.schemes.projective.projective_space.ProjectiveSpace_field>`.

    INPUT:

    - ``A`` -- ambient :class:`projective space
      <sage.schemes.projective.projective_space.ProjectiveSpace_field>`

    - ``polynomials`` -- single polynomial, ideal or iterable of
      defining homogeneous polynomials

    EXAMPLES::

        sage: P.<x, y, z> = ProjectiveSpace(2, QQ)
        sage: P.subscheme([x^2 - y*z])
        Closed subscheme of Projective Space of dimension 2 over Rational Field defined by:
          x^2 - y*z

    TESTS::

        sage: from sage.schemes.projective.projective_subscheme import AlgebraicScheme_subscheme_projective
        sage: AlgebraicScheme_subscheme_projective(P, [x^2-y*z])
        Closed subscheme of Projective Space of dimension 2 over Rational Field defined by:
          x^2 - y*z
    """
    def point(self, v, check=True):
        """
        Create a point on this projective subscheme.

        INPUT:

        - ``v`` -- anything that defines a point

        - ``check`` -- boolean (default: ``True``); whether
          to check the defining data for consistency

        OUTPUT: a point of the subscheme

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x - y, y - z])
            sage: X.point([1,1,1])
            (1 : 1 : 1)

        ::

            sage: P2.<x,y> = ProjectiveSpace(QQ, 1)
            sage: X = P2.subscheme([y])
            sage: X.point(infinity)
            (1 : 0)

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: X = P.subscheme(x^2 + 2*y^2)
            sage: X.point(infinity)
            Traceback (most recent call last):
            ...
            TypeError: Coordinates [1, 0] do not define a point on Closed subscheme
            of Projective Space of dimension 1 over Rational Field defined by:
              x^2 + 2*y^2
        """
        from sage.rings.infinity import infinity
        if v is infinity or\
           (isinstance(v, (list, tuple)) and len(v) == 1 and v[0] is infinity):
            if self.ambient_space().dimension_relative() > 1:
                raise ValueError("%s not well defined in dimension > 1" % v)
            v = [1, 0]
        # todo: update elliptic curve stuff to take point_homset as argument
        if isinstance(self, EllipticCurve):
            try:
                return self._point(self.point_homset(), v, check=check)
            except AttributeError:  # legacy code without point_homset
                return self._point(self, v, check=check)

        return self.point_homset()(v, check=check)

    def _morphism(self, *args, **kwds):
        r"""
        Construct a morphism determined by action on points of ``self``.

        For internal use only.

        INPUT:

        - same as for
          :class:`~sage.schemes.projective.projective_morphism.SchemeMorphism_polynomial_projective_space`.

        OUTPUT: :class:`~sage.schemes.projective.projective_morphism.SchemeMorphism_polynomial_projective_space`

        TESTS::

            sage: P1.<x,y> = ProjectiveSpace(1, QQ)
            sage: P2 = ProjectiveSpace(2, QQ)
            sage: H12 = P1.Hom(P2)
            sage: H12([x^2, x*y, y^2])    # indirect doctest
            Scheme morphism:
              From: Projective Space of dimension 1 over Rational Field
              To:   Projective Space of dimension 2 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to (x^2 : x*y : y^2)
            sage: P1._morphism(H12, [x^2, x*y, y^2])
            Scheme morphism:
              From: Projective Space of dimension 1 over Rational Field
              To:   Projective Space of dimension 2 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to (x^2 : x*y : y^2)
        """
        return self.ambient_space()._morphism(*args, **kwds)

    def dimension(self):
        """
        Return the dimension of the projective algebraic subscheme.

        OUTPUT: integer

        EXAMPLES::

            sage: # needs sage.libs.singular
            sage: P2.<x,y,z> = ProjectiveSpace(2, QQ)
            sage: P2.subscheme([]).dimension()
            2
            sage: P2.subscheme([x]).dimension()
            1
            sage: P2.subscheme([x^5]).dimension()
            1
            sage: P2.subscheme([x^2 + y^2 - z^2]).dimension()
            1
            sage: P2.subscheme([x*(x-z), y*(y-z)]).dimension()
            0

        Something less obvious::

            sage: P3.<x,y,z,w,t> = ProjectiveSpace(4, QQ)
            sage: X = P3.subscheme([x^2, x^2*y^2 + z^2*t^2,
            ....:                   z^2 - w^2, 10*x^2 + w^2 - z^2]); X
            Closed subscheme of Projective Space of dimension 4 over Rational Field defined by:
              x^2,
              x^2*y^2 + z^2*t^2,
              z^2 - w^2,
              10*x^2 - z^2 + w^2
            sage: X.dimension()                                                         # needs sage.libs.singular
            1
        """
        try:
            return self.__dimension
        except AttributeError:
            self.__dimension = self.defining_ideal().dimension() - 1
            return self.__dimension

    def affine_patch(self, i, AA=None):
        r"""
        Return the `i`-th affine patch of this projective scheme.

        This is the intersection with this `i`-th affine patch of
        its ambient space.

        INPUT:

        - ``i`` -- integer between 0 and dimension of ``self``, inclusive

        - ``AA`` -- (default: ``None``) ambient affine space, this
          is constructed if it is not given

        OUTPUT:

        An affine algebraic scheme with fixed
        :meth:`embedding_morphism` equal to the default
        :meth:`projective_embedding` map`.

        EXAMPLES::

            sage: # needs sage.libs.singular
            sage: PP = ProjectiveSpace(2, QQ, names='X,Y,Z')
            sage: X,Y,Z = PP.gens()
            sage: C = PP.subscheme(X^3*Y + Y^3*Z + Z^3*X)
            sage: U = C.affine_patch(0); U
            Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
              Y^3*Z + Z^3 + Y
            sage: U.embedding_morphism()
            Scheme morphism:
              From: Closed subscheme of Affine Space of dimension 2 over Rational Field
                    defined by: Y^3*Z + Z^3 + Y
              To:   Closed subscheme of Projective Space of dimension 2 over Rational Field
                    defined by: X^3*Y + Y^3*Z + X*Z^3
              Defn: Defined on coordinates by sending (Y, Z) to (1 : Y : Z)
            sage: U.projective_embedding() is U.embedding_morphism()
            True

        ::

            sage: A.<x,y,z> = AffineSpace(QQ, 3)
            sage: X = A.subscheme([x - y*z])
            sage: Y = X.projective_embedding(1).codomain()                              # needs sage.libs.singular
            sage: Y.affine_patch(1, A).ambient_space() == A                             # needs sage.libs.singular
            True

        ::

            sage: P.<u,v,w> = ProjectiveSpace(2, ZZ)
            sage: S = P.subscheme([u^2 - v*w])
            sage: A.<x, y> = AffineSpace(2, ZZ)
            sage: S.affine_patch(1, A)                                                  # needs sage.libs.singular
            Closed subscheme of Affine Space of dimension 2 over Integer Ring defined by:
              x^2 - y
        """
        i = int(i)   # implicit type checking
        PP = self.ambient_space()
        n = PP.dimension_relative()
        if i < 0 or i > n:
            raise ValueError(f"Argument i (= {i}) must be between 0 and {n}.")
        try:
            A = self.__affine_patches[i]
            # assume that if you've passed in a new ambient affine space
            # you want to override the existing patch
            if AA is None or A.ambient_space() == AA:
                return self.__affine_patches[i]
        except AttributeError:
            self.__affine_patches = {}
        except KeyError:
            pass
        if AA is None:
            AA = PP.affine_patch(i)
        elif AA.dimension_relative() != n:
            raise ValueError("Affine Space must be of the dimension %s" % (n))
        phi = AA.projective_embedding(i, PP)
        polys = self.defining_polynomials()
        xi = phi.defining_polynomials()
        U = AA.subscheme([f(xi) for f in polys])
        phi = U.projective_embedding(i, PP)
        U._embedding_morphism = phi
        self.__affine_patches[i] = U
        return U

    def _best_affine_patch(self, point):
        r"""
        Return the best affine patch of the ambient projective space.

        The "best" affine patch is where you end up dividing by the
        homogeneous coordinate with the largest absolute
        value. Division by small numbers is numerically unstable.

        INPUT:

        - ``point`` -- a point of the algebraic subscheme

        OUTPUT: integer. The index of the patch. See :meth:`affine_patch`

        EXAMPLES::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: S = P.subscheme(x + 2*y + 3*z)
            sage: S._best_affine_patch(P.point([0,-3,2]))
            1
            sage: S._best_affine_patch([0,-3,2])
            1

        TESTS::

            sage: F = GF(3)
            sage: P.<x,y,z> = ProjectiveSpace(F, 2)
            sage: S._best_affine_patch([0,1,2])
            2
        """
        point = list(point)
        try:
            abs_point = [abs(c) for c in point]
        except ArithmeticError:
            # our base ring does not know abs
            abs_point = point
        # find best patch
        i_max = 0
        p_max = abs_point[i_max]
        for i in range(1, len(point)):
            if abs_point[i] > p_max:
                i_max = i
                p_max = abs_point[i_max]
        return i_max

    def neighborhood(self, point):
        r"""
        Return an affine algebraic subscheme isomorphic to a
        neighborhood of the ``point``.

        INPUT:

        - ``point`` -- a point of the projective subscheme

        OUTPUT:

        An affine algebraic scheme (polynomial equations in affine
        space) ``result`` such that

        * :meth:`embedding_morphism
          <AlgebraicScheme.embedding_morphism>` is an isomorphism to a
          neighborhood of ``point``

        * :meth:`embedding_center <AlgebraicScheme.embedding_center>`
          is mapped to ``point``.

        EXAMPLES::

            sage: P.<x,y,z>= ProjectiveSpace(QQ, 2)
            sage: S = P.subscheme(x + 2*y + 3*z)
            sage: s = S.point([0,-3,2]); s
            (0 : -3/2 : 1)
            sage: patch = S.neighborhood(s); patch
            Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
              x + 3*z
            sage: patch.embedding_morphism()
            Scheme morphism:
              From: Closed subscheme of Affine Space of dimension 2 over Rational Field
                    defined by: x + 3*z
              To:   Closed subscheme of Projective Space of dimension 2 over Rational Field
                    defined by: x + 2*y + 3*z
              Defn: Defined on coordinates by sending (x, z) to (x : -3/2 : z + 1)
            sage: patch.embedding_center()
            (0, 0)
            sage: patch.embedding_morphism()([0,0])
            (0 : -3/2 : 1)
            sage: patch.embedding_morphism()(patch.embedding_center())
            (0 : -3/2 : 1)
        """
        point = list(point)
        self._check_satisfies_equations(point)
        PP = self.ambient_space()
        n = PP.dimension()
        i = self._best_affine_patch(point)

        patch_cover = PP.affine_patch(i)
        R = patch_cover.coordinate_ring()

        phi = list(point)
        for j in range(i):
            phi[j] += R.gen(j)
        for j in range(i, n):
            phi[j + 1] += R.gen(j)

        pullback_polys = [f(phi) for f in self.defining_polynomials()]
        return patch_cover.subscheme(pullback_polys, embedding_center=[0] * n,
                                     embedding_codomain=self,
                                     embedding_images=phi)

    def is_smooth(self, point=None) -> bool:
        r"""
        Test whether the algebraic subscheme is smooth.

        INPUT:

        - ``point`` -- a point or ``None`` (default); the point to
          test smoothness at

        OUTPUT:

        boolean; if no point was specified, returns whether the
        algebraic subscheme is smooth everywhere. Otherwise,
        smoothness at the specified point is tested.

        EXAMPLES::

            sage: # needs sage.libs.singular
            sage: P2.<x,y,z> = ProjectiveSpace(2, QQ)
            sage: cuspidal_curve = P2.subscheme([y^2*z - x^3])
            sage: cuspidal_curve
            Closed subscheme of Projective Space of dimension 2 over Rational Field defined by:
              -x^3 + y^2*z
            sage: cuspidal_curve.is_smooth([1,1,1])
            True
            sage: cuspidal_curve.is_smooth([0,0,1])
            False
            sage: cuspidal_curve.is_smooth()
            False
            sage: P2.subscheme([y^2*z - x^3 + z^3 + 1/10*x*y*z]).is_smooth()
            True

        TESTS::

            sage: H = P2.subscheme(x)                                                   # needs sage.libs.singular
            sage: H.is_smooth()  # one of the few cases where the cone over the subvariety is smooth                    # needs sage.libs.singular
            True
        """
        if point is not None:
            self._check_satisfies_equations(point)
            R = self.ambient_space().coordinate_ring()
            point_subs = dict(zip(R.gens(), point))
            Jac = self.Jacobian().subs(point_subs)
            return not Jac.is_zero()

        # testing smoothness everywhere tends to be expensive
        try:
            return self._smooth
        except AttributeError:
            pass
        sing_dim = self.Jacobian().dimension()
        # We really test the affine cone here; the origin is always a
        # singular point:
        self._smooth = (sing_dim <= 0)
        return self._smooth

    def orbit(self, f, N) -> list:
        r"""
        Return the orbit of this scheme by ``f``.

        If `N` is an integer it returns `[self,f(self),\ldots,f^N(self)]`.
        If `N` is a list or tuple `N=[m,k]` it returns `[f^m(self),\ldots,f^k(self)`].

        INPUT:

        - ``f`` -- a :class:`DynamicalSystem_projective` with ``self`` in ``f.domain()``

        - ``N`` -- nonnegative integer or list or tuple of two nonnegative integers

        OUTPUT: list of projective subschemes

        EXAMPLES::

            sage: # needs sage.libs.singular sage.schemes
            sage: P.<x,y,z,w> = ProjectiveSpace(QQ, 3)
            sage: f = DynamicalSystem_projective([(x-2*y)^2, (x-2*z)^2,
            ....:                                 (x-2*w)^2, x^2])
            sage: f.orbit(P.subscheme([x]), 5)
            [Closed subscheme of Projective Space of dimension 3 over Rational Field
              defined by: x,
             Closed subscheme of Projective Space of dimension 3 over Rational Field
              defined by: w,
             Closed subscheme of Projective Space of dimension 3 over Rational Field
              defined by: z - w,
             Closed subscheme of Projective Space of dimension 3 over Rational Field
              defined by: y - z,
             Closed subscheme of Projective Space of dimension 3 over Rational Field
              defined by: x - y,
             Closed subscheme of Projective Space of dimension 3 over Rational Field
              defined by: x - w]

        ::

            sage: PS.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: P1.<u,v> = ProjectiveSpace(QQ, 1)
            sage: H = Hom(PS, P1)
            sage: f = H([x^2, y^2])
            sage: X = PS.subscheme([x - y])
            sage: X.orbit(f, 2)
            Traceback (most recent call last):
            ...
            TypeError: map must be a dynamical system for iteration

        ::

            sage: PS.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: f = DynamicalSystem_projective([x^2, y^2, z^2])                       # needs sage.schemes
            sage: X = PS.subscheme([x - y])
            sage: X.orbit(f, [-1,2])                                                    # needs sage.schemes
            Traceback (most recent call last):
            ...
            TypeError: orbit bounds must be nonnegative
        """
        if not isinstance(f, DynamicalSystem):
            raise TypeError("map must be a dynamical system for iteration")
        if not isinstance(N, (list, tuple)):
            N = [0, N]
        N[0] = ZZ(N[0])
        N[1] = ZZ(N[1])
        if N[0] < 0 or N[1] < 0:
            raise TypeError("orbit bounds must be nonnegative")
        if N[0] > N[1]:
            return []

        Q = self
        for i in range(1, N[0] + 1):
            Q = f(Q)
        Orb = [Q]

        for i in range(N[0] + 1, N[1] + 1):
            Q = f(Q)
            Orb.append(Q)
        return Orb

    def nth_iterate(self, f, n):
        r"""
        The `n`-th forward image of this scheme by the map ``f``.

        INPUT:

        - ``f`` -- a :class:`DynamicalSystem_projective` with ``self`` in ``f.domain()``

        - ``n`` -- positive integer

        OUTPUT:

        - A subscheme in ``f.codomain()``

        EXAMPLES::

            sage: P.<x,y,z,w> = ProjectiveSpace(QQ, 3)
            sage: f = DynamicalSystem_projective([y^2, z^2, x^2, w^2])                  # needs sage.schemes
            sage: f.nth_iterate(P.subscheme([x - w, y - z]), 3)                         # needs sage.libs.singular sage.schemes
            Closed subscheme of Projective Space of dimension 3 over Rational Field
             defined by:
              y - z,
              x - w

        ::

            sage: PS.<x,y,z> = ProjectiveSpace(ZZ, 2)
            sage: f = DynamicalSystem_projective([x^2, y^2, z^2])                       # needs sage.schemes
            sage: X = PS.subscheme([x - y])
            sage: X.nth_iterate(f, -2)                                                  # needs sage.libs.singular sage.schemes
            Traceback (most recent call last):
            ...
            TypeError: must be a forward orbit

        ::

            sage: PS.<x,y,z> = ProjectiveSpace(ZZ, 2)
            sage: P2.<u,v,w> = ProjectiveSpace(QQ, 2)
            sage: H = Hom(PS, P2)
            sage: f = H([x^2, y^2, z^2])
            sage: X = PS.subscheme([x - y])
            sage: X.nth_iterate(f, 2)
            Traceback (most recent call last):
            ...
            TypeError: map must be a dynamical system for iteration

        ::

            sage: PS.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: f = DynamicalSystem_projective([x^2, y^2, z^2])                       # needs sage.schemes
            sage: X = PS.subscheme([x - y])
            sage: X.nth_iterate(f, 2.5)                                                 # needs sage.schemes
            Traceback (most recent call last):
            ...
            TypeError: Attempt to coerce non-integral RealNumber to Integer
        """
        n = ZZ(n)
        if n < 0:
            raise TypeError("must be a forward orbit")
        return self.orbit(f, [n, n + 1])[0]

    def _forward_image(self, f, check=True):
        r"""
        Compute the forward image of this subscheme by the morphism ``f``.

        The forward image is computed through elimination and ``f`` must be
        a morphism for this to be well defined.
        In particular, let `X = V(h_1,\ldots, h_t)` and define the ideal
        `I = (h_1,\ldots,h_t,y_0-f_0(\bar{x}), \ldots, y_n-f_n(\bar{x}))`.
        Then the elimination ideal `I_{n+1} = I \cap K[y_0,\ldots,y_n]` is a homogeneous
        ideal and `self(X) = V(I_{n+1})`.

        INPUT:

        - ``f`` -- a map whose domain contains ``self``

        - ``check`` -- boolean, if ``False`` no input checking is done

        OUTPUT: a subscheme in the codomain of ``f``

        EXAMPLES::

            sage: PS.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: H = End(PS)
            sage: f = H([x^2, y^2 - 2*z^2, z^2])
            sage: X = PS.subscheme(y - 2*z)
            sage: X._forward_image(f)                                                   # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 2 over Rational Field
             defined by:
              y - 2*z

        ::

            sage: set_verbose(None)
            sage: PS.<x,y,z,w> = ProjectiveSpace(ZZ, 3)
            sage: H = End(PS)
            sage: f = H([y^2, x^2, w^2, z^2])
            sage: X = PS.subscheme([z^2 + y*w, x - w])
            sage: f(X)                                                                  # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 3 over Integer Ring
             defined by:
              y - z,
              x*z - w^2

        ::

            sage: # needs sage.rings.real_mpfr
            sage: PS.<x,y,z,w> = ProjectiveSpace(CC, 3)
            sage: H = End(PS)
            sage: f = H([x^2 + y^2, y^2, z^2-y^2, w^2])
            sage: X = PS.subscheme([z - 2*w])
            sage: f(X)                                                                  # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 3 over Complex Field
             with 53 bits of precision defined by:
              y + z + (-4.00000000000000)*w

        ::

            sage: R.<t> = PolynomialRing(QQ)
            sage: P.<x,y,z> = ProjectiveSpace(FractionField(R), 2)
            sage: H = End(P)
            sage: f = H([x^2 + 2*y*z, t^2*y^2, z^2])
            sage: f([t^2*y - z])                                                        # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 2 over Fraction Field
             of Univariate Polynomial Ring in t over Rational Field defined by:
              y - 1/(t^2)*z

        ::

            sage: # needs sage.rings.padics
            sage: set_verbose(-1)
            sage: PS.<x,y,z> = ProjectiveSpace(Qp(3), 2)
            sage: H = End(PS)
            sage: f = H([x^2, 2*y^2, z^2])
            sage: X = PS.subscheme([2*x - y, z])
            sage: f(X)                                                                  # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 2 over 3-adic Field
             with capped relative precision 20 defined by:
              z,
              x + (1 + 3^2 + 3^4 + 3^6 + 3^8 + 3^10 + 3^12 + 3^14 + 3^16 + 3^18 + O(3^20))*y

        ::

            sage: R.<y0,y1,y2,y3> = PolynomialRing(QQ)
            sage: P.<x,y,z> = ProjectiveSpace(FractionField(R), 2)
            sage: H = End(P)
            sage: f = H([y0*x^2 + y1*z^2, y2*y^2 + y3*z^2, z^2])
            sage: X = P.subscheme(x*z)
            sage: X._forward_image(f)                                                   # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 2 over Fraction Field
             of Multivariate Polynomial Ring in y0, y1, y2, y3 over Rational Field
             defined by:
              x*z + (-y1)*z^2

            ::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: P5.<z0,z1,z2,z3,z4,z5> = ProjectiveSpace(QQ, 5)
            sage: H = Hom(P2, P5)
            sage: f = H([x^2, x*y, x*z, y^2, y*z, z^2])  # Veronese map
            sage: X = P2.subscheme([])
            sage: f(X)                                                                  # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 5 over Rational Field
             defined by:
              -z4^2 + z3*z5,
              -z2*z4 + z1*z5,
              -z2*z3 + z1*z4,
              -z2^2 + z0*z5,
              -z1*z2 + z0*z4,
              -z1^2 + z0*z3

            ::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: P3.<u,v,w,t> = ProjectiveSpace(QQ, 3)
            sage: H = Hom(P2, P3)
            sage: X = P2.subscheme([x - y, x - z])
            sage: f = H([x^2, y^2, z^2, x*y])
            sage: f(X)                                                                  # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 3 over Rational Field
             defined by:
              w - t,
              v - t,
              u - t

            ::

            sage: P1.<u,v> = ProjectiveSpace(QQ, 1)
            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: H = Hom(P2, P1)
            sage: f = H([x^2, y*z])
            sage: X = P2.subscheme([x - y])
            sage: f(X)                                                                  # needs sage.libs.singular
            Traceback (most recent call last):
            ...
            TypeError: map must be a morphism

            ::

            sage: PS.<x,y,z> = ProjectiveSpace(ZZ, 2)
            sage: H = End(PS)
            sage: f = H([x^3, x*y^2, x*z^2])
            sage: X = PS.subscheme([x - y])
            sage: X._forward_image(f)                                                   # needs sage.libs.singular
            Traceback (most recent call last):
            ...
            TypeError: map must be a morphism

        ::

            sage: PS.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: P1.<u,v> = ProjectiveSpace(QQ, 1)
            sage: Y = P1.subscheme([u - v])
            sage: H = End(PS)
            sage: f = H([x^2, y^2, z^2])
            sage: Y._forward_image(f)                                                   # needs sage.libs.singular
            Traceback (most recent call last):
            ...
            TypeError: subscheme must be in ambient space of domain of map
        """
        if check:
            if not f.is_morphism():
                raise TypeError("map must be a morphism")
        dom = f.domain()
        codom = f.codomain()
        if check:
            if self.ambient_space() != dom:
                raise TypeError("subscheme must be in ambient space of domain of map")
        CR_dom = dom.coordinate_ring()
        CR_codom = codom.coordinate_ring()
        n = CR_dom.ngens()
        m = CR_codom.ngens()
        # can't call eliminate if the base ring is polynomial so we do it ourselves
        # with a lex ordering
        R = PolynomialRing(f.base_ring(), n + m, 'tempvar', order='lex')
        Rvars = R.gens()[0:n]
        phi = CR_dom.hom(Rvars, R)
        zero = n * [0]
        psi = R.hom(zero + list(CR_codom.gens()), CR_codom)
        # set up ideal
        L = R.ideal([phi(t) for t in self.defining_polynomials()] + [R.gen(n + i) - phi(f[i]) for i in range(m)])
        G = L.groebner_basis()  # eliminate
        newL = []
        # get only the elimination ideal portion
        for i in range(len(G) - 1, 0, -1):
            v = G[i].variables()
            if all(Rvars[j] not in v for j in range(n)):
                newL.append(psi(G[i]))
        return codom.subscheme(newL)

    def preimage(self, f, k=1, check=True):
        r"""
        The subscheme that maps to this scheme by the map `f^k`.

        In particular, `f^{-k}(V(h_1,\ldots,h_t)) = V(h_1 \circ f^k, \ldots, h_t \circ f^k)`.
        Map must be a morphism and also must be an endomorphism for `k > 1`.

        INPUT:

        - ``f`` -- a map whose codomain contains this scheme

        - ``k`` -- positive integer

        - ``check`` -- boolean; if ``False`` no input checking is done

        OUTPUT: a subscheme in the domain of ``f``

        EXAMPLES::

            sage: PS.<x,y,z> = ProjectiveSpace(ZZ, 2)
            sage: H = End(PS)
            sage: f = H([y^2, x^2, z^2])
            sage: X = PS.subscheme([x - y])
            sage: X.preimage(f)                                                         # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 2 over Integer Ring
             defined by:
              -x^2 + y^2

        ::

            sage: P.<x,y,z,w,t> = ProjectiveSpace(QQ, 4)
            sage: H = End(P)
            sage: f = H([x^2 - y^2, y^2, z^2, w^2, t^2 + w^2])
            sage: f.rational_preimages(P.subscheme([x - z, t^2, w - t]))                # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 4 over Rational Field
             defined by:
              x^2 - y^2 - z^2,
              w^4 + 2*w^2*t^2 + t^4,
              -t^2

        ::

            sage: P1.<x,y> = ProjectiveSpace(QQ, 1)
            sage: P3.<u,v,w,t> = ProjectiveSpace(QQ, 3)
            sage: H = Hom(P1, P3)
            sage: X = P3.subscheme([u - v, 2*u - w, u + t])
            sage: f = H([x^2, y^2, x^2 + y^2, x*y])
            sage: X.preimage(f)                                                         # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 1 over Rational Field
             defined by:
              x^2 - y^2,
              x^2 - y^2,
              x^2 + x*y

        ::

            sage: P1.<x,y> = ProjectiveSpace(QQ, 1)
            sage: P3.<u,v,w,t> = ProjectiveSpace(QQ, 3)
            sage: H = Hom(P3, P1)
            sage: X = P1.subscheme([x - y])
            sage: f = H([u^2, v^2])
            sage: X.preimage(f)                                                         # needs sage.libs.singular
            Traceback (most recent call last):
            ...
            TypeError: map must be a morphism

        ::

            sage: PS.<x,y,z> = ProjectiveSpace(ZZ, 2)
            sage: H = End(PS)
            sage: f = H([x^2, x^2, x^2])
            sage: X = PS.subscheme([x - y])
            sage: X.preimage(f)                                                         # needs sage.libs.singular
            Traceback (most recent call last):
            ...
            TypeError: map must be a morphism

        ::

            sage: PS.<x,y,z> = ProjectiveSpace(ZZ, 2)
            sage: P1.<u,v> = ProjectiveSpace(ZZ, 1)
            sage: Y = P1.subscheme([u^2 - v^2])
            sage: H = End(PS)
            sage: f = H([x^2, y^2, z^2])
            sage: Y.preimage(f)                                                         # needs sage.libs.singular
            Traceback (most recent call last):
            ...
            TypeError: subscheme must be in ambient space of codomain

        ::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: Y = P.subscheme([x - y])
            sage: H = End(P)
            sage: f = H([x^2, y^2, z^2])
            sage: Y.preimage(f, k=2)                                                    # needs sage.libs.singular sage.schemes
            Closed subscheme of Projective Space of dimension 2 over Rational Field
             defined by:
              x^4 - y^4
        """
        dom = f.domain()
        codom = f.codomain()
        if check:
            if not f.is_morphism():
                raise TypeError("map must be a morphism")
            if self.ambient_space() != codom:
                raise TypeError("subscheme must be in ambient space of codomain")
            k = ZZ(k)
            if k <= 0:
                raise ValueError("k (=%s) must be a positive integer" % (k))
            if k > 1 and not f.is_endomorphism():
                raise TypeError("map must be an endomorphism")
        R = codom.coordinate_ring()
        F = f.as_dynamical_system().nth_iterate_map(k) if k > 1 else f
        dic = {R.gen(i): F[i] for i in range(codom.dimension_relative() + 1)}
        return dom.subscheme([t.subs(dic) for t in self.defining_polynomials()])

    def dual(self):
        r"""
        Return the projective dual of the given subscheme of projective space.

        INPUT:

        - ``X`` -- a subscheme of projective space. At present, ``X`` is
          required to be an irreducible and reduced hypersurface defined
          over `\QQ` or a finite field.

        OUTPUT: the dual of ``X`` as a subscheme of the dual projective space

        EXAMPLES:

        The dual of a smooth conic in the plane is also a smooth conic::

            sage: R.<x, y, z> = QQ[]
            sage: P.<x, y, z> = ProjectiveSpace(2, QQ)
            sage: I = R.ideal(x^2 + y^2 + z^2)
            sage: X = P.subscheme(I)
            sage: X.dual()                                                              # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 2 over Rational Field
             defined by:
              y0^2 + y1^2 + y2^2

        The dual of the twisted cubic curve in projective 3-space is a singular
        quartic surface. In the following example, we compute the dual of this
        surface, which by double duality is equal to the twisted cubic itself.
        The output is the twisted cubic as an intersection of three quadrics::

            sage: R.<x, y, z, w> = QQ[]
            sage: P.<x, y, z, w> = ProjectiveSpace(3, QQ)
            sage: I = R.ideal(y^2*z^2 - 4*x*z^3 - 4*y^3*w + 18*x*y*z*w - 27*x^2*w^2)
            sage: X = P.subscheme(I)
            sage: X.dual()                                                              # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 3 over
             Rational Field defined by:
              y2^2 - y1*y3,
              y1*y2 - y0*y3,
              y1^2 - y0*y2

        The singular locus of the quartic surface in the last example
        is itself supported on a twisted cubic::

            sage: X.Jacobian().radical()                                                # needs sage.libs.singular
            Ideal (z^2 - 3*y*w, y*z - 9*x*w, y^2 - 3*x*z) of Multivariate
            Polynomial Ring in x, y, z, w over Rational Field

        An example over a finite field::

            sage: R = PolynomialRing(GF(61), 'a,b,c')
            sage: P.<a, b, c> = ProjectiveSpace(2, R.base_ring())
            sage: X = P.subscheme(R.ideal(a*a + 2*b*b + 3*c*c))
            sage: X.dual()                                                              # needs sage.libs.singular sage.rings.finite_rings
            Closed subscheme of Projective Space of dimension 2 over
             Finite Field of size 61 defined by:
              y0^2 - 30*y1^2 - 20*y2^2

        TESTS::

            sage: # needs sage.rings.padics
            sage: R = PolynomialRing(Qp(3), 'a,b,c')
            sage: P.<a, b, c> = ProjectiveSpace(2, R.base_ring())
            sage: X = P.subscheme(R.ideal(a*a + 2*b*b + 3*c*c))
            sage: X.dual()
            Traceback (most recent call last):
            ...
            NotImplementedError: base ring must be QQ or a finite field
        """
        from sage.libs.singular.function_factory import ff

        K = self.base_ring()
        if not (isinstance(K, RationalField) or K in Fields().Finite()):
            raise NotImplementedError("base ring must be QQ or a finite field")
        I = self.defining_ideal()
        m = I.ngens()
        n = I.ring().ngens() - 1
        if (m != 1 or (n < 1) or I.is_zero()
                or I.is_trivial() or not I.is_prime()):
            raise NotImplementedError("At the present, the method is only"
                                      " implemented for irreducible and"
                                      " reduced hypersurfaces and the given"
                                      " list of generators for the ideal must"
                                      " have exactly one element.")
        R = PolynomialRing(K, 'x', n + 1)
        from sage.schemes.projective.projective_space import ProjectiveSpace
        Pd = ProjectiveSpace(n, K, 'y')
        Rd = Pd.coordinate_ring()
        x = R.variable_names()
        y = Rd.variable_names()
        S = PolynomialRing(K, x + y + ('t',))
        if S.has_coerce_map_from(I.ring()):
            T = PolynomialRing(K, 'w', n + 1)
            I_S = (I.change_ring(T)).change_ring(S)
        else:
            I_S = I.change_ring(S)
        f_S = I_S.gens()[0]
        z = S.gens()
        J = I_S
        for i in range(n + 1):
            J = J + S.ideal(z[-1] * f_S.derivative(z[i]) - z[i + n + 1])

        try:
            sat = ff.elim__lib.sat_with_exp
        except NameError:
            sat = ff.elim__lib.sat

        max_ideal = S.ideal(z[n + 1: 2 * n + 2])
        J_sat_gens = sat(J, max_ideal)[0]
        J_sat = S.ideal(J_sat_gens)
        L = J_sat.elimination_ideal(z[0: n + 1] + (z[-1],))
        return Pd.subscheme(L.change_ring(Rd))

    def degree(self):
        r"""
        Return the degree of this projective subscheme.

        If `P(t) = a_{m}t^m + \ldots + a_{0}` is the Hilbert
        polynomial of this subscheme, then the degree is `a_{m} m!`.

        OUTPUT: integer

        EXAMPLES::

            sage: P.<x,y,z,w,t,u> = ProjectiveSpace(QQ, 5)
            sage: X = P.subscheme([x^7 + x*y*z*t^4 - u^7])
            sage: X.degree()                                                            # needs sage.libs.singular
            7

            sage: P.<x,y,z,w> = ProjectiveSpace(GF(13), 3)
            sage: X = P.subscheme([y^3 - w^3, x + 7*z])
            sage: X.degree()                                                            # needs sage.libs.singular
            3

            sage: # needs sage.libs.singular sage.schemes
            sage: P.<x,y,z,w,u> = ProjectiveSpace(QQ, 4)
            sage: C = P.curve([x^7 - y*z^3*w^2*u, w*x^2 - y*u^2, z^3 + y^3])
            sage: C.degree()
            63
        """
        P = self.defining_ideal().hilbert_polynomial()
        return P.leading_coefficient() * P.degree().factorial()

    def intersection_multiplicity(self, X, P):
        r"""
        Return the intersection multiplicity of this subscheme and the subscheme ``X`` at the point ``P``.

        This uses the intersection_multiplicity function for affine subschemes on affine patches of this subscheme
        and ``X`` that contain ``P``.

        INPUT:

        - ``X`` -- subscheme in the same ambient space as this subscheme

        - ``P`` -- a point in the intersection of this subscheme with ``X``

        OUTPUT: integer

        EXAMPLES::

            sage: # needs sage.schemes
            sage: P.<x,y,z> = ProjectiveSpace(GF(5), 2)
            sage: C = Curve([x^4 - z^2*y^2], P)
            sage: D = Curve([y^4*z - x^5 - x^3*z^2], P)
            sage: Q1 = P([0,1,0])
            sage: C.intersection_multiplicity(D, Q1)                                    # needs sage.libs.singular
            4
            sage: Q2 = P([0,0,1])
            sage: C.intersection_multiplicity(D, Q2)                                    # needs sage.libs.singular
            6

        ::

            sage: # needs sage.rings.number_field
            sage: R.<a> = QQ[]
            sage: K.<b> = NumberField(a^4 + 1)
            sage: P.<x,y,z,w> = ProjectiveSpace(K, 3)
            sage: X = P.subscheme([x^2 + y^2 - z*w])
            sage: Y = P.subscheme([y*z - x*w, z - w])
            sage: Q1 = P([b^2,1,0,0])
            sage: X.intersection_multiplicity(Y, Q1)                                    # needs sage.libs.singular
            1
            sage: Q2 = P([1/2*b^3 - 1/2*b, 1/2*b^3 - 1/2*b, 1, 1])
            sage: X.intersection_multiplicity(Y, Q2)                                    # needs sage.libs.singular
            1

        ::

            sage: P.<x,y,z,w> = ProjectiveSpace(QQ, 3)
            sage: X = P.subscheme([x^2 - z^2, y^3 - w*x^2])
            sage: Y = P.subscheme([w^2 - 2*x*y + z^2, y^2 - w^2])
            sage: Q = P([1,1,-1,1])
            sage: X.intersection_multiplicity(Y, Q)                                     # needs sage.libs.singular
            Traceback (most recent call last):
            ...
            TypeError: the intersection of this subscheme and (=Closed subscheme of Affine Space of dimension 3
            over Rational Field defined by: z^2 + w^2 - 2*y, y^2 - w^2)
            must be proper and finite
        """
        try:
            self.ambient_space()(P)
        except TypeError:
            raise TypeError("(={}) must be a point in the ambient space of this subscheme and (={})".format(P, X))
        # find an affine chart of the ambient space of this curve that contains P
        n = self.ambient_space().dimension_relative()
        for i in range(n + 1):
            if P[i] != 0:
                break
        X1 = self.affine_patch(i)
        X2 = X.affine_patch(i)
        return X1.intersection_multiplicity(X2, X1(P.dehomogenize(i)))

    def multiplicity(self, P):
        r"""
        Return the multiplicity of ``P`` on this subscheme.

        This is computed as the multiplicity of the corresponding point on an affine patch of this subscheme
        that contains ``P``. This subscheme must be defined over a field. An error is returned if ``P``
        not a point on this subscheme.

        INPUT:

        - ``P`` -- a point on this subscheme

        OUTPUT: integer

        EXAMPLES::

            sage: P.<x,y,z,w,t> = ProjectiveSpace(QQ, 4)
            sage: X = P.subscheme([y^2 - x*t, w^7 - t*w*x^5 - z^7])
            sage: Q1 = P([0,0,1,1,1])
            sage: X.multiplicity(Q1)                                                    # needs sage.libs.singular
            1
            sage: Q2 = P([1,0,0,0,0])
            sage: X.multiplicity(Q2)                                                    # needs sage.libs.singular
            3
            sage: Q3 = P([0,0,0,0,1])
            sage: X.multiplicity(Q3)                                                    # needs sage.libs.singular
            7

        ::

            sage: # needs sage.rings.real_mpfr
            sage: P.<x,y,z,w> = ProjectiveSpace(CC, 3)
            sage: X = P.subscheme([z^5*x^2*w - y^8])
            sage: Q = P([2,0,0,1])
            sage: X.multiplicity(Q)                                                     # needs sage.libs.singular
            5

        ::

            sage: # needs sage.libs.singular sage.schemes
            sage: P.<x,y,z,w> = ProjectiveSpace(GF(29), 3)
            sage: C = Curve([y^17 - x^5*w^4*z^8, x*y - z^2], P)
            sage: Q = P([3,0,0,1])
            sage: C.multiplicity(Q)
            8
        """
        if self.base_ring() not in Fields():
            raise TypeError("subscheme must be defined over a field")

        # check whether P is a point on this subscheme
        try:
            P = self(P)
        except TypeError:
            raise TypeError(f"(={P}) is not a point on (={self})")

        # find an affine chart of the ambient space of self that contains P
        i = 0
        while P[i] == 0:
            i += 1
        X = self.affine_patch(i)
        return X.multiplicity(X(P.dehomogenize(i)))

    def veronese_embedding(self, d, CS=None, order='lex'):
        r"""
        Return the degree ``d`` Veronese embedding of this projective subscheme.

        INPUT:

        - ``d`` -- positive integer

        - ``CS`` -- (default: ``None``) a projective ambient space to embed
          into. If the projective ambient space of this subscheme is of
          dimension `N`, the dimension of ``CS`` must be
          `\binom{N + d}{d} - 1`. This is constructed if not specified.

        - ``order`` -- string (default: ``'lex'``); a monomial order to use to
          arrange the monomials defining the embedding. The monomials will be
          arranged from greatest to least with respect to this order.

        OUTPUT: a scheme morphism from this subscheme to its image by the
        degree ``d`` Veronese embedding

        EXAMPLES::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: L = P.subscheme([y - x])
            sage: v = L.veronese_embedding(2); v                                        # needs sage.libs.singular
            Scheme morphism:
              From: Closed subscheme of Projective Space of dimension 2
                    over Rational Field defined by: -x + y
              To:   Closed subscheme of Projective Space of dimension 5
                    over Rational Field defined by:
                      -x4^2 + x3*x5,
                      x2 - x4,
                      x1 - x3,
                      x0 - x3
              Defn: Defined on coordinates by sending (x : y : z) to
                    (x^2 : x*y : x*z : y^2 : y*z : z^2)
            sage: v.codomain().degree()                                                 # needs sage.libs.singular
            2
            sage: C = P.subscheme([y*z - x^2])
            sage: C.veronese_embedding(2).codomain().degree()                           # needs sage.libs.singular
            4

        twisted cubic::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: Q.<u,v,s,t> = ProjectiveSpace(QQ, 3)
            sage: P.subscheme([]).veronese_embedding(3, Q)                              # needs sage.libs.singular
            Scheme morphism:
              From: Closed subscheme of Projective Space of dimension 1
                    over Rational Field defined by: (no polynomials)
              To:   Closed subscheme of Projective Space of dimension 3
                    over Rational Field defined by:
                      -s^2 + v*t,
                      -v*s + u*t,
                      -v^2 + u*s
              Defn: Defined on coordinates by sending (x : y) to
                    (x^3 : x^2*y : x*y^2 : y^3)
        """
        # construct map between projective spaces
        v = self.ambient_space().veronese_embedding(d, CS, order)
        # return this map restricted to self and its image
        return Hom(self, v(self))(v.defining_polynomials())


class AlgebraicScheme_subscheme_projective_field(AlgebraicScheme_subscheme_projective):
    """
    Algebraic subschemes of projective spaces defined over fields.
    """
    def _morphism(self, *args, **kwds):
        r"""
        Construct a morphism determined by action on points of ``self``.

        For internal use only.

        INPUT:

        - same as for
          :class:`~sage.schemes.projective.projective_morphism.SchemeMorphism_polynomial_projective_space`.

        OUTPUT: :class:`~sage.schemes.projective.projective_morphism.SchemeMorphism_polynomial_projective_space`

        TESTS::

            sage: P1.<x,y> = ProjectiveSpace(1, QQ)
            sage: P2 = ProjectiveSpace(2, QQ)
            sage: H12 = P1.Hom(P2)
            sage: H12([x^2, x*y, y^2])    # indirect doctest
            Scheme morphism:
              From: Projective Space of dimension 1 over Rational Field
              To:   Projective Space of dimension 2 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to (x^2 : x*y : y^2)
            sage: P1._morphism(H12, [x^2, x*y, y^2])
            Scheme morphism:
              From: Projective Space of dimension 1 over Rational Field
              To:   Projective Space of dimension 2 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to (x^2 : x*y : y^2)
        """
        return SchemeMorphism_polynomial_projective_subscheme_field(*args, **kwds)

    def Chow_form(self):
        r"""
        Return the Chow form associated to this subscheme.

        For a `k`-dimensional subvariety of `\mathbb{P}^N` of degree `D`.
        The `(N-k-1)`-dimensional projective linear subspaces of `\mathbb{P}^N`
        meeting `X` form a hypersurface in the Grassmannian `G(N-k-1,N)`.
        The homogeneous form of degree `D` defining this hypersurface in Plucker
        coordinates is called the Chow form of `X`.

        The base ring needs to be a number field, finite field, or `\QQbar`.

        ALGORITHM:

        For a `k`-dimension subscheme `X` consider the `k+1` linear forms
        `l_i = u_{i0}x_0 + \cdots + u_{in}x_n`. Let `J` be the ideal in the
        polynomial ring `K[x_i,u_{ij}]` defined by the equations of `X` and the `l_i`.
        Let `J'` be the saturation of `J` with respect to the irrelevant ideal of
        the ambient projective space of `X`. The elimination ideal `I = J' \cap K[u_{ij}]`
        is a principal ideal, let `R` be its generator. The Chow form is obtained by
        writing `R` as a polynomial in Plucker coordinates (i.e. bracket polynomials).
        [DS1994]_.

        OUTPUT: a homogeneous polynomial

        EXAMPLES::

            sage: P.<x0,x1,x2,x3> = ProjectiveSpace(GF(17), 3)
            sage: X = P.subscheme([x3 + x1, x2 - x0, x2 - x3])
            sage: X.Chow_form()                                                         # needs sage.libs.singular
            t0 - t1 + t2 + t3

        ::

            sage: P.<x0,x1,x2,x3> = ProjectiveSpace(QQ, 3)
            sage: X = P.subscheme([x3^2 - 101*x1^2 - 3*x2*x0])
            sage: X.Chow_form()                                                         # needs sage.libs.singular
            t0^2 - 101*t2^2 - 3*t1*t3

        ::

            sage: # needs sage.libs.singular
            sage: P.<x0,x1,x2,x3> = ProjectiveSpace(QQ, 3)
            sage: X = P.subscheme([x0*x2 - x1^2, x0*x3 - x1*x2, x1*x3 - x2^2])
            sage: Ch = X.Chow_form(); Ch
            t2^3 + 2*t2^2*t3 + t2*t3^2 - 3*t1*t2*t4 - t1*t3*t4 + t0*t4^2 + t1^2*t5
            sage: Y = P.subscheme_from_Chow_form(Ch, 1); Y
            Closed subscheme of Projective Space of dimension 3 over Rational Field
             defined by:
              x2^2*x3 - x1*x3^2,                    -x2^3 + x0*x3^2,
              -x2^2*x3 + x1*x3^2,                   x1*x2*x3 - x0*x3^2,
              3*x1*x2^2 - 3*x0*x2*x3,               -2*x1^2*x3 + 2*x0*x2*x3,
              -3*x1^2*x2 + 3*x0*x1*x3,              x1^3 - x0^2*x3,
              x2^3 - x1*x2*x3,                      -3*x1*x2^2 + 2*x1^2*x3 + x0*x2*x3,
              2*x0*x2^2 - 2*x0*x1*x3,               3*x1^2*x2 - 2*x0*x2^2 - x0*x1*x3,
              -x0*x1*x2 + x0^2*x3,                  -x0*x1^2 + x0^2*x2,
              -x1^3 + x0*x1*x2,                     x0*x1^2 - x0^2*x2
            sage: I = Y.defining_ideal()
            sage: I.saturation(I.ring().ideal(list(I.ring().gens())))[0]
            Ideal (x2^2 - x1*x3, x1*x2 - x0*x3, x1^2 - x0*x2)
             of Multivariate Polynomial Ring in x0, x1, x2, x3 over Rational Field
        """
        I = self.defining_ideal()
        P = self.ambient_space()
        R = P.coordinate_ring()
        N = P.dimension() + 1
        d = self.dimension()
        # create the ring for the generic linear hyperplanes
        # u0x0 + u1x1 + ...
        SS = PolynomialRing(R.base_ring(), 'u', N * (d + 1), order='lex')
        vars = SS.variable_names() + R.variable_names()
        S = PolynomialRing(R.base_ring(), vars, order='lex')
        n = S.ngens()
        newcoords = [S.gen(n - N + t) for t in range(N)]
        # map the generators of the subscheme into the ring with the hyperplane variables
        phi = R.hom(newcoords, S)
        phi(self.defining_polynomials()[0])
        # create the dim(X)+1 linear hyperplanes
        l = []
        for i in range(d + 1):
            t = 0
            for j in range(N):
                t += S.gen(N * i + j) * newcoords[j]
            l.append(t)
        # intersect the hyperplanes with X
        J = phi(I) + S.ideal(l)
        # saturate the ideal with respect to the irrelevant ideal
        J2 = J.saturation(S.ideal([phi(u) for u in R.gens()]))[0]
        # eliminate the original variables to be left with the hyperplane coefficients 'u'
        E = J2.elimination_ideal(newcoords)
        # create the plucker coordinates
        D = binomial(N, N - d - 1)  # number of plucker coordinates
        tvars = [f't{i}' for i in range(D)]  # plucker coordinates
        T = PolynomialRing(R.base_ring(), tvars + list(S.variable_names()), order='lex')
        L = []
        coeffs = [T.gen(i) for i in range(len(tvars), N * (d + 1) + len(tvars))]
        M = matrix(T, d + 1, N, coeffs)
        i = 0
        for c in M.minors(d + 1):
            L.append(T.gen(i) - c)
            i += 1
        # create the ideal that we can use for eliminating to get a polynomial
        # in the plucker coordinates (brackets)
        br = T.ideal(L)
        # create a mapping into a polynomial ring over the plucker coordinates
        # and the hyperplane coefficients
        psi = S.hom(coeffs + [0 for _ in range(N)], T)
        E2 = T.ideal([psi(u) for u in E.gens()] + br)
        # eliminate the hyperplane coefficients
        CH = E2.elimination_ideal(coeffs)
        # CH should be a principal ideal, but because of the relations among
        # the plucker coordinates, the elimination will probably have several generators

        # get the relations among the plucker coordinates
        rel = br.elimination_ideal(coeffs)
        # reduce CH with respect to the relations
        reduced = [f.reduce(rel) for f in CH.gens()]
        # find the principal generator

        # polynomial ring in just the plucker coordinates
        T2 = PolynomialRing(R.base_ring(), tvars)
        alp = T.hom(tvars + (N * (d + 1) + N) * [0], T2)
        # get the degrees of the reduced generators of CH
        degs = [u.degree() for u in reduced]
        mind = max(degs)
        # need the smallest degree form that did not reduce to 0
        for d in degs:
            if d < mind and d > 0:
                mind = d
        ind = degs.index(mind)
        CF = reduced[ind]  # this should be the Chow form of X
        # check that it is correct (i.e., it is a principal generator for CH + the relations)
        rel2 = rel + [CF]
        assert all(f in rel2 for f in CH.gens()), "did not find a principal generator"
        return alp(CF)

    def global_height(self, prec=None):
        """
        Return the (projective) global height of the subscheme.

        INPUT:

        - ``prec`` -- desired floating point precision (default:
          default ``RealField`` precision)

        OUTPUT: a real number

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: NF.<a> = NumberField(x^2 - 5)
            sage: P.<x,y,z> = ProjectiveSpace(NF, 2)
            sage: X = P.subscheme([x^2 + y*z, 2*y*z, 3*x*y])
            sage: X.global_height()                                                     # needs sage.libs.singular
            0.000000000000000

        ::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P.subscheme([z^2 - 101*y^2 - 3*x*z])
            sage: X.global_height()             # long time                             # needs sage.libs.singular
            4.61512051684126
        """
        return self.Chow_form().global_height(prec)

    def local_height(self, v, prec=None):
        """
        Return the (projective) local height of the subscheme.

        INPUT:

        - ``v`` -- a prime or prime ideal of the base ring

        - ``prec`` -- desired floating point precision (default:
          default ``RealField`` precision)

        OUTPUT: a real number

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: NF.<a> = NumberField(x^2 - 5)
            sage: I = NF.ideal(3)
            sage: P.<x,y,z> = ProjectiveSpace(NF, 2)
            sage: X = P.subscheme([3*x*y - 5*x*z, y^2])
            sage: X.local_height(I)                                                     # needs sage.libs.singular
            0.000000000000000

        ::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P.subscheme([z^2 - 101*y^2 - 3*x*z])
            sage: X.local_height(2)                                                     # needs sage.libs.singular
            0.000000000000000
        """
        return self.Chow_form().local_height(v, prec)

    def local_height_arch(self, i, prec=None):
        """
        Return the local height at the ``i``-th infinite place of the subscheme.

        INPUT:

        - ``i`` -- integer

        - ``prec`` -- desired floating point precision (default:
          default ``RealField`` precision)

        OUTPUT: a real number

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: NF.<a> = NumberField(x^2 - 5)
            sage: P.<x,y,z> = ProjectiveSpace(NF, 2)
            sage: X = P.subscheme([x^2 + y*z, 3*x*y])
            sage: X.local_height_arch(1)                                                # needs sage.libs.singular
            0.0000000000000000000000000000000

        ::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P.subscheme([z^2 - 101*y^2 - 3*x*z])
            sage: X.local_height_arch(1)                                                # needs sage.libs.singular
            4.61512051684126
        """
        return self.Chow_form().local_height_arch(i, prec)



# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.rings.number_field
"""
Base class of number fields

AUTHORS:

- William Stein (2007-09-04): initial version
"""


def is_NumberField(x):
    """
    Return ``True`` if ``x`` is of number field type.

    This function is deprecated.

    EXAMPLES::

        sage: from sage.rings.number_field.number_field_base import is_NumberField
        sage: x = polygen(ZZ)
        sage: is_NumberField(NumberField(x^2 + 1, 'a'))
        doctest:...: DeprecationWarning: the function is_NumberField is deprecated; use
        isinstance(x, sage.rings.number_field.number_field_base.NumberField) instead
        See https://github.com/sagemath/sage/issues/35283 for details.
        True
        sage: is_NumberField(QuadraticField(-97, 'theta'))
        True
        sage: is_NumberField(CyclotomicField(97))
        True

    Note that the rational numbers ``QQ`` are a number field.::

        sage: is_NumberField(QQ)
        True
        sage: is_NumberField(ZZ)
        False
    """
    from sage.misc.superseded import deprecation
    deprecation(35283,
                "the function is_NumberField is deprecated; use "
                "isinstance(x, sage.rings.number_field.number_field_base.NumberField) instead")

    return isinstance(x, NumberField)


from sage.rings.ring cimport Field


cdef class NumberField(Field):
    r"""
    Base class for all number fields.

    TESTS::

        sage: x = polygen(ZZ)
        sage: k = NumberField(x^2 + 1, 'i')
        sage: k == loads(dumps(k))
        True

        sage: z = polygen(QQ)
        sage: K.<theta, beta> = NumberField([z^3 - 3, z^2 + 1])
        sage: K.is_finite()
        False
        sage: K.order()
        +Infinity
    """
    # This token docstring is mostly there to prevent Sphinx from pasting in
    # the docstring of the __init__ method inherited from IntegralDomain, which
    # is rather confusing.
    def _pushout_(self, other):
        r"""
        If ``self`` and/or ``other`` are embedded, use this embedding to
        discover a common parent.

        Currently embeddings into ``AA`` and ``QQbar`` are supported.

        TESTS:

        Pushout is implemented for number field embedded in ``AA``::

            sage: x = polygen(ZZ)
            sage: K.<a> = NumberField(x^2 - 3, embedding=AA(3)**(1/2))
            sage: L.<b> = NumberField(x^2 - 2, embedding=AA(2)**(1/2))
            sage: cm = sage.structure.element.get_coercion_model()
            sage: cm.explain(K,L,operator.add)
            Coercion on left operand via
                Generic morphism:
                  From: Number Field in a with defining polynomial x^2 - 3 with a = 1.732050807568878?
                  To:   Algebraic Real Field
                  Defn: a -> 1.732050807568878?
            Coercion on right operand via
                Generic morphism:
                  From: Number Field in b with defining polynomial x^2 - 2 with b = 1.414213562373095?
                  To:   Algebraic Real Field
                  Defn: b -> 1.414213562373095?
            Arithmetic performed after coercions.
            Result lives in Algebraic Real Field
            Algebraic Real Field

        As a consequence, operations and comparisons work nicely::

            sage: a + b
            3.146264369941973?
            sage: a < b
            False
            sage: 3*a < 4*b
            True

        Using number field with other classes::

            sage: K.<cbrt2> = NumberField(x^3 - 2, embedding=AA(2)**(1/3))
            sage: (cbrt2 + a) * b
            4.231287179063857?
            sage: b + QQbar(-3).sqrt()
            1.414213562373095? + 1.732050807568878?*I

        Pushout is implemented for number field embedded in ``QQbar``::

            sage: Km2.<sqrtm2> = NumberField(x^2 + 2, embedding=QQbar(-2).sqrt())
            sage: b + sqrtm2
            1.414213562373095? + 1.414213562373095?*I
            sage: sqrtm2 + b
            1.414213562373095? + 1.414213562373095?*I
            sage: sqrtm2 + AA(3).sqrt()
            1.732050807568878? + 1.414213562373095?*I
        """
        # Use the embedding of ``self``, if it exists.
        if self._embedding:
            codomain_self = self._embedding.codomain()
        else:
            codomain_self = self

        # Use the embedding of ``other``, if it exists.
        if isinstance(other, NumberField):
            embedding = (<NumberField>other)._embedding
            if embedding:
                codomain_other = embedding.codomain()
            else:
                codomain_other = other
        else:
            codomain_other = other

        try:
            from sage.rings.qqbar import AA, QQbar
        except ImportError:
            pass
        else:
            if codomain_self is AA and codomain_other is AA:
                return AA
            if codomain_self in (AA, QQbar) and codomain_other in (AA, QQbar):
                return QQbar

    def ring_of_integers(self, *args, **kwds):
        r"""
        Synonym for :meth:`maximal_order`.

        EXAMPLES::

            sage: x = polygen(ZZ)
            sage: K.<a> = NumberField(x^2 + 1)
            sage: K.ring_of_integers()
            Gaussian Integers generated by a in Number Field in a with defining polynomial x^2 + 1
        """
        return self.maximal_order(*args, **kwds)

    def OK(self, *args, **kwds):
        r"""
        Synonym for :meth:`maximal_order`.

        EXAMPLES::

            sage: x = polygen(ZZ)
            sage: NumberField(x^3 - 2,'a').OK()
            Maximal Order generated by a in Number Field in a with defining polynomial x^3 - 2
        """
        return self.maximal_order(*args, **kwds)

    def maximal_order(self):
        r"""
        Return the maximal order, i.e., the ring of integers of this
        number field.

        EXAMPLES::

            sage: x = polygen(ZZ)
            sage: NumberField(x^3 - 2,'b').maximal_order()
            Maximal Order generated by b in Number Field in b with defining polynomial x^3 - 2
        """
        raise NotImplementedError

    def is_absolute(self):
        r"""
        Return ``True`` if ``self`` is viewed as a single extension over `\QQ`.

        EXAMPLES::

            sage: x = polygen(ZZ)
            sage: K.<a> = NumberField(x^3 + 2)
            sage: K.is_absolute()
            True
            sage: y = polygen(K)
            sage: L.<b> = NumberField(y^2 + 1)
            sage: L.is_absolute()
            False
            sage: QQ.is_absolute()
            True
        """
        raise NotImplementedError

    def signature(self):
        r"""
        Return `(r_1, r_2)`, where `r_1` and `r_2` are the number of real embeddings
        and pairs of complex embeddings of this field, respectively.

        EXAMPLES::

            sage: x = polygen(ZZ)
            sage: NumberField(x^3 - 2, 'a').signature()
            (1, 1)
        """
        raise NotImplementedError

    def degree(self):
        """
        Return the degree of this number field.

        EXAMPLES::

            sage: x = polygen(ZZ)
            sage: NumberField(x^3 + 9, 'a').degree()
            3
        """
        raise NotImplementedError

    def discriminant(self):
        """
        Return the discriminant of this number field.

        EXAMPLES::

            sage: x = polygen(ZZ)
            sage: NumberField(x^3 + 9, 'a').discriminant()
            -243
        """
        raise NotImplementedError

    def minkowski_bound(self):
        r"""
        Return the Minkowski bound associated to this number field.

        This is a bound `B` so that every integral ideal is equivalent
        modulo principal fractional ideals to an integral ideal of
        norm at most `B`.

        .. SEEALSO::

            :meth:`~bach_bound`

        OUTPUT: symbolic expression or Rational

        EXAMPLES:

        The Minkowski bound for `\QQ[i]` tells us that the class
        number is 1::

            sage: # needs sage.symbolic
            sage: K = QQ[I]
            sage: B = K.minkowski_bound(); B
            4/pi
            sage: B.n()
            1.27323954473516

        We compute the Minkowski bound for `\QQ[\sqrt[3]{2}]`::

            sage: # needs sage.symbolic
            sage: K = QQ[2^(1/3)]
            sage: B = K.minkowski_bound(); B
            16/3*sqrt(3)/pi
            sage: B.n()
            2.94042077558289
            sage: int(B)
            2

        We compute the Minkowski bound for `\QQ[\sqrt{10}]`, which has class
        number 2::

            sage: # needs sage.symbolic
            sage: K = QQ[sqrt(10)]
            sage: B = K.minkowski_bound(); B
            sqrt(10)
            sage: int(B)
            3
            sage: K.class_number()
            2

        We compute the Minkowski bound for `\QQ[\sqrt{2}+\sqrt{3}]`::

            sage: # needs sage.symbolic
            sage: x = polygen(ZZ)
            sage: K.<y,z> = NumberField([x^2 - 2, x^2 - 3])
            sage: L.<w> = QQ[sqrt(2) + sqrt(3)]
            sage: B = K.minkowski_bound(); B
            9/2
            sage: int(B)
            4
            sage: B == L.minkowski_bound()
            True
            sage: K.class_number()
            1

        The bound of course also works for the rational numbers::

            sage: QQ.minkowski_bound()
            1
        """
        _, s = self.signature()
        n = self.absolute_degree()
        d = self.absolute_discriminant().abs().sqrt()
        from sage.symbolic.constants import pi
        if s > 0:
            return d * (4/pi)**s * n.factorial() / (n**n)
        else:
            return d * n.factorial() / (n**n)

    def bach_bound(self):
        r"""
        Return the Bach bound associated to this number field.

        Assuming the General Riemann Hypothesis, this is a bound `B` so
        that every integral ideal is equivalent modulo principal
        fractional ideals to an integral ideal of norm at most `B`.

        .. SEEALSO::

            :meth:`~minkowski_bound`

        OUTPUT: symbolic expression or the Integer 1

        EXAMPLES:

        We compute both the Minkowski and Bach bounds for a quadratic
        field, where the Minkowski bound is much better::

            sage: # needs sage.symbolic
            sage: K = QQ[sqrt(5)]
            sage: K.minkowski_bound()
            1/2*sqrt(5)
            sage: K.minkowski_bound().n()
            1.11803398874989
            sage: K.bach_bound()
            12*log(5)^2
            sage: K.bach_bound().n()
            31.0834847277628

        We compute both the Minkowski and Bach bounds for a bigger
        degree field, where the Bach bound is much better::

            sage: # needs sage.symbolic
            sage: K = CyclotomicField(37)
            sage: K.minkowski_bound().n()
            7.50857335698544e14
            sage: K.bach_bound().n()
            191669.304126267

        The bound of course also works for the rational numbers::

            sage: QQ.bach_bound()                                                       # needs sage.symbolic
            1
        """
        ans = 12 * abs(self.discriminant()).log()**2
        if ans == 0: # rational numbers
            from sage.rings.integer import Integer
            return Integer(1)
        return ans

    # Approximate embeddings for comparisons with respect to the order of RR or
    # CC

    def _init_embedding_approx(self):
        r"""
        Initialize the approximation of embeddings.

        This should be called only once.

        TESTS::

            sage: x = polygen(ZZ)
            sage: K.<a> = NumberField(x^3 - x^2 - x - 1, embedding=1)
            sage: K._get_embedding_approx(0)   # indirect doctest
            1.839286755214161?
        """

        if self._gen_approx is not None or self._embedding is None:
            return

        try:
            from sage.rings.qqbar import AA
        except ImportError:
            AA = None

        try:
            from sage.rings.real_lazy import RLF
        except ImportError:
            RLF = None

        codomain = self._embedding.codomain()
        if codomain is AA or codomain is RLF:
            self._gen_approx = []
            self._embedded_real = 1

    cpdef _get_embedding_approx(self, size_t i):
        r"""
        Return an interval approximation of the generator of this number field.

        OUTPUT: a real interval element with precision `53 \times 2^i`

        EXAMPLES::

            sage: x = polygen(ZZ)
            sage: p = x^5 - 3*x + 1
            sage: a_AA = AA.polynomial_root(p, RIF(0,1))
            sage: K.<a> = NumberField(p, embedding=a_AA)
            sage: K._get_embedding_approx(2)
            0.3347341419433526870750989624732833071257517550374285560578335863?
            sage: K._get_embedding_approx(1)
            0.33473414194335268707509896247329?
            sage: K._get_embedding_approx(1).str(style='brackets')
            '[0.334734141943352687075098962473280 .. 0.334734141943352687075098962473287]'

            sage: K._get_embedding_approx(2).prec()
            212
            sage: K._get_embedding_approx(1).prec()
            106
            sage: K._get_embedding_approx(0).prec()
            53

        If a real embedding is not specified, this method will result in an error::

            sage: N.<g> = NumberField(x^3 + 2)
            sage: N._get_embedding_approx(1)
            Traceback (most recent call last):
            ...
            ValueError: No embedding set. You need to specify a real embedding.


        .. SEEALSO::

            :class:` RealIntervalField_class <sage.rings.real_mpfi.RealIntervalField_class>`
        """
        if self._embedded_real and i < len(self._gen_approx):
            return self._gen_approx[i]

        cdef size_t j
        if self._embedded_real:
            j = len(self._gen_approx)
            from sage.rings.real_mpfi import RealIntervalField
            gen = self._embedding.gen_image()
            while j <= i:
                self._gen_approx.append(RealIntervalField(53 << j)(gen))
                j += 1
            return self._gen_approx[i]
        else:
            raise ValueError("No embedding set. You need to specify a real embedding.")

    def _matrix_charpoly(self, M, var):
        r"""
        Use PARI to compute the characteristic polynomial of ``self`` as a
        polynomial over the base ring.

        EXAMPLES::

            sage: x = QQ['x'].gen()
            sage: K.<a> = NumberField(x^2 - 2)
            sage: m = matrix(K, [[a-1, 2], [a, a+1]])
            sage: m.charpoly('Z')   # indirect doctest
            Z^2 - 2*a*Z - 2*a + 1
            sage: m.charpoly('a')(m) == 0   # indirect doctest
            True
            sage: m = matrix(K, [[0, a, 0], [-a, 0, 0], [0, 0, 0]])
            sage: m.charpoly('Z')   # indirect doctest
            Z^3 + 2*Z

         ::

            sage: L.<b> = K.extension(x^3 - a)
            sage: m = matrix(L, [[b+a, 1], [a, b^2-2]])
            sage: m.charpoly('Z')   # indirect doctest
            Z^2 + (-b^2 - b - a + 2)*Z + a*b^2 - 2*b - 2*a
            sage: m.charpoly('a')   # indirect doctest
            a^2 + (-b^2 - b - a + 2)*a + a*b^2 - 2*b - 2*a
            sage: m.charpoly('a')(m) == 0
            True

        ::

            sage: M.<c> = L.extension(x^2 - a*x + b)
            sage: m = matrix(M, [[a+b+c, 0, b], [0, c, 1], [a-1, b^2+1, 2]])
            sage: f = m.charpoly('Z'); f    # indirect doctest
            Z^3 + (-2*c - b - a - 2)*Z^2 + ((b + 2*a + 4)*c - b^2 + (-a + 2)*b + 2*a - 1)*Z + (b^2 + (a - 3)*b - 4*a + 1)*c + a*b^2 + 3*b + 2*a
            sage: f(m) == 0
            True
            sage: f.base_ring() is M
            True
        """
        paripoly = M.__pari__().charpoly()
        return self[var](paripoly)



# sage_setup: distribution = sagemath-categories
r"""
Trigonometric functions
"""
import math

from sage.symbolic.function import GinacFunction


class Function_sin(GinacFunction):
    def __init__(self):
        r"""
        The sine function.

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: sin(0)
            0
            sage: sin(x).subs(x==0)
            0
            sage: sin(2).n(100)
            0.90929742682568169539601986591
            sage: sin(x)._sympy_()                                                      # needs sympy
            sin(x)

        We can prevent evaluation using the ``hold`` parameter::

            sage: sin(0, hold=True)                                                     # needs sage.symbolic
            sin(0)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: a = sin(0, hold=True); a.simplify()                                   # needs sage.symbolic
            0

        If possible, the argument is also reduced modulo the
        period length `2\pi`, and well-known identities are
        directly evaluated::

            sage: k = var('k', domain='integer')                                        # needs sage.symbolic
            sage: sin(1 + 2*k*pi)                                                       # needs sage.symbolic
            sin(1)
            sage: sin(k*pi)                                                             # needs sage.symbolic
            0

        TESTS::

            sage: loads(dumps(sin))
            sin

            sage: conjugate(sin(x))                                                     # needs sage.symbolic
            sin(conjugate(x))
            sage: sin(complex(1,1))     # rel tol 1e-15
            (1.2984575814159773+0.6349639147847361j)

            sage: # needs sage.symbolic
            sage: sin(pi/5)
            1/4*sqrt(-2*sqrt(5) + 10)
            sage: sin(pi/8)
            1/2*sqrt(-sqrt(2) + 2)
            sage: sin(pi/24)
            1/4*sqrt(-2*sqrt(6) - 2*sqrt(2) + 8)
            sage: sin(pi/30)
            -1/8*sqrt(5) + 1/4*sqrt(-3/2*sqrt(5) + 15/2) - 1/8
            sage: sin(104*pi/105)
            sin(1/105*pi)
            sage: cos(pi/8)
            1/2*sqrt(sqrt(2) + 2)
            sage: cos(pi/10)
            1/4*sqrt(2*sqrt(5) + 10)
            sage: cos(pi/12)
            1/4*sqrt(6) + 1/4*sqrt(2)
            sage: cos(pi/15)
            1/8*sqrt(5) + 1/4*sqrt(3/2*sqrt(5) + 15/2) - 1/8
            sage: cos(pi/24)
            1/4*sqrt(2*sqrt(6) + 2*sqrt(2) + 8)
            sage: cos(104*pi/105)
            -cos(1/105*pi)
            sage: tan(pi/5)
            sqrt(-2*sqrt(5) + 5)
            sage: tan(pi/8)
            sqrt(2) - 1
            sage: tan(pi/10)
            1/5*sqrt(-10*sqrt(5) + 25)
            sage: tan(pi/16)
            -sqrt(2) + sqrt(2*sqrt(2) + 4) - 1
            sage: tan(pi/20)
            sqrt(5) - sqrt(2*sqrt(5) + 5) + 1
            sage: tan(pi/24)
            sqrt(6) - sqrt(3) + sqrt(2) - 2
            sage: tan(104*pi/105)
            -tan(1/105*pi)
            sage: cot(104*pi/105)
            -cot(1/105*pi)
            sage: sec(104*pi/105)
            -sec(1/105*pi)
            sage: csc(104*pi/105)
            csc(1/105*pi)

            sage: all(sin(rat*pi).n(200) - sin(rat*pi, hold=True).n(200) < 1e-30        # needs sage.symbolic
            ....:     for rat in [1/5, 2/5, 1/30, 7/30, 11/30, 13/30,
            ....:                 1/8, 3/8, 1/24, 5/24, 7/24, 11/24])
            True
            sage: all(cos(rat*pi).n(200)-cos(rat*pi, hold=True).n(200) < 1e-30          # needs sage.symbolic
            ....:     for rat in [1/10, 3/10, 1/12, 5/12, 1/15, 2/15, 4/15, 7/15,
            ....:                 1/8, 3/8, 1/24, 5/24, 11/24])
            True
            sage: all(tan(rat*pi).n(200)-tan(rat*pi, hold=True).n(200) < 1e-30          # needs sage.symbolic
            ....:     for rat in [1/5, 2/5, 1/10, 3/10, 1/20, 3/20, 7/20, 9/20,
            ....:                 1/8, 3/8, 1/16, 3/16, 5/16, 7/16, 1/24, 5/24, 7/24, 11/24])
            True

        Check that :issue:`20456` is fixed::

            sage: assume(x > 0)                                                         # needs sage.symbolic
            sage: sin(pi*x)                                                             # needs sage.symbolic
            sin(pi*x)
            sage: forget()                                                              # needs sage.symbolic

        Check that :issue:`20752` is fixed::

            sage: sin(3*pi + 41/42*pi)                                                  # needs sage.symbolic
            -sin(1/42*pi)
            sage: sin(-5*pi + 1/42*pi)                                                  # needs sage.symbolic
            -sin(1/42*pi)
            sage: sin(pi - 1/42*pi)                                                     # needs sage.symbolic
            sin(1/42*pi)
        """
        GinacFunction.__init__(self, 'sin', latex_name=r"\sin",
                conversions=dict(maxima='sin', mathematica='Sin',
                                 giac='sin', fricas='sin', sympy='sin'))


sin = Function_sin()


class Function_cos(GinacFunction):
    def __init__(self):
        r"""
        The cosine function.

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: cos(pi)
            -1
            sage: cos(x).subs(x==pi)
            -1
            sage: cos(2).n(100)
            -0.41614683654714238699756822950
            sage: cos(x)._sympy_()                                                      # needs sympy
            cos(x)

        We can prevent evaluation using the ``hold`` parameter::

            sage: cos(0, hold=True)                                                     # needs sage.symbolic
            cos(0)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: a = cos(0, hold=True); a.simplify()                                   # needs sage.symbolic
            1

        If possible, the argument is also reduced modulo the
        period length `2\pi`, and well-known identities are
        directly evaluated::

            sage: # needs sage.symbolic
            sage: k = var('k', domain='integer')
            sage: cos(1 + 2*k*pi)
            cos(1)
            sage: cos(k*pi)
            cos(pi*k)
            sage: cos(pi/3 + 2*k*pi)
            1/2

        TESTS::

            sage: loads(dumps(cos))
            cos

            sage: conjugate(cos(x))                                                     # needs sage.symbolic
            cos(conjugate(x))
            sage: cos(complex(1,1))     # rel tol 1e-15
            (0.8337300251311491-0.9888977057628651j)

        Check that :issue:`20752` is fixed::

            sage: cos(3*pi + 41/42*pi)                                                  # needs sage.symbolic
            cos(1/42*pi)
            sage: cos(-5*pi + 1/42*pi)                                                  # needs sage.symbolic
            -cos(1/42*pi)
            sage: cos(pi - 1/42*pi)                                                     # needs sage.symbolic
            -cos(1/42*pi)
        """
        GinacFunction.__init__(self, 'cos', latex_name=r"\cos",
                conversions=dict(maxima='cos', mathematica='Cos',
                                 giac='cos', fricas='cos', sympy='cos'))


cos = Function_cos()


class Function_tan(GinacFunction):
    def __init__(self):
        r"""
        The tangent function.

        EXAMPLES::

            sage: # needs sage.rings.real_mpfr
            sage: tan(3.1415)
            -0.0000926535900581913
            sage: tan(3.1415/4)
            0.999953674278156

            sage: # needs sage.symbolic
            sage: tan(pi)
            0
            sage: tan(pi/4)
            1
            sage: tan(1/2)
            tan(1/2)
            sage: RR(tan(1/2))
            0.546302489843790

        We can prevent evaluation using the ``hold`` parameter::

            sage: tan(pi/4, hold=True)                                                  # needs sage.symbolic
            tan(1/4*pi)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: a = tan(pi/4, hold=True); a.simplify()                                # needs sage.symbolic
            1

        If possible, the argument is also reduced modulo the
        period length `\pi`, and well-known identities are
        directly evaluated::

            sage: k = var('k', domain='integer')                                        # needs sage.symbolic
            sage: tan(1 + 2*k*pi)                                                       # needs sage.symbolic
            tan(1)
            sage: tan(k*pi)                                                             # needs sage.symbolic
            0

        TESTS::

            sage: tan(x)._sympy_()                                                      # needs sympy sage.symbolic
            tan(x)
            sage: conjugate(tan(x))                                                     # needs sage.symbolic
            tan(conjugate(x))
            sage: tan(complex(1,1))     # rel tol 1e-15
            (0.2717525853195118+1.0839233273386946j)

        Check that :issue:`19791` is fixed::

            sage: tan(2+I).imag().n()                                                   # needs sage.symbolic
            1.16673625724092
        """
        GinacFunction.__init__(self, 'tan', latex_name=r"\tan",
                               conversions=dict(maxima='tan', mathematica='Tan',
                                 giac='tan', fricas='tan', sympy='tan'))


tan = Function_tan()


class Function_cot(GinacFunction):
    def __init__(self):
        r"""
        The cotangent function.

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: cot(pi/4)
            1
            sage: RR(cot(pi/4))
            1.00000000000000
            sage: cot(1/2)
            cot(1/2)
            sage: cot(0.5)
            1.83048772171245

            sage: latex(cot(x))                                                         # needs sage.symbolic
            \cot\left(x\right)
            sage: cot(x)._sympy_()                                                      # needs sympy sage.symbolic
            cot(x)

        We can prevent evaluation using the ``hold`` parameter::

            sage: cot(pi/4, hold=True)                                                  # needs sage.symbolic
            cot(1/4*pi)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: a = cot(pi/4, hold=True); a.simplify()                                # needs sage.symbolic
            1

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: cot(pi/4)
            1
            sage: cot(x).subs(x==pi/4)
            1
            sage: cot(pi/7)
            cot(1/7*pi)
            sage: cot(x)
            cot(x)

            sage: # needs sage.symbolic
            sage: n(cot(pi/4), 100)
            1.0000000000000000000000000000
            sage: float(cot(1))
            0.64209261593433...
            sage: bool(diff(cot(x), x) == diff(1/tan(x), x))
            True
            sage: diff(cot(x), x)
            -cot(x)^2 - 1

        TESTS::

            sage: # needs sage.symbolic
            sage: cot(float(0))
            Infinity
            sage: cot(SR(0))
            Infinity
            sage: cot(float(0.1))
            9.966644423259238
            sage: type(_)
            <... 'float'>
            sage: cot(float(0))
            Infinity
            sage: cot(SR(0))
            Infinity
            sage: cot(float(0.1))
            9.966644423259238
            sage: type(_)
            <... 'float'>

        Test complex input::

            sage: cot(complex(1,1))     # rel tol 1e-15                                 # needs sage.rings.complex_double
            (0.21762156185440273-0.8680141428959249j)
            sage: cot(1.+I)                                                             # needs sage.symbolic
            0.217621561854403 - 0.868014142895925*I
        """
        GinacFunction.__init__(self, 'cot', latex_name=r"\cot",
                               conversions=dict(maxima='cot', mathematica='Cot',
                                 giac='cot', fricas='cot', sympy='cot'))

    def _eval_numpy_(self, x):
        """
        EXAMPLES::

             sage: import numpy                                                         # needs numpy
             sage: a = numpy.arange(2, 5)                                               # needs numpy
             sage: cot(a)                                                               # needs numpy
             array([-0.45765755, -7.01525255,  0.86369115])
        """
        return 1.0 / tan(x)


cot = Function_cot()


class Function_sec(GinacFunction):
    def __init__(self):
        r"""
        The secant function.

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: sec(pi/4)
            sqrt(2)
            sage: sec(x).subs(x==pi/4)
            sqrt(2)
            sage: sec(pi/7)
            sec(1/7*pi)
            sage: sec(x)
            sec(x)
            sage: RR(sec(pi/4))
            1.41421356237310
            sage: n(sec(pi/4),100)
            1.4142135623730950488016887242
            sage: float(sec(pi/4))
            1.4142135623730951
            sage: sec(1/2)
            sec(1/2)
            sage: sec(0.5)
            1.13949392732455

            sage: # needs sage.symbolic
            sage: bool(diff(sec(x), x) == diff(1/cos(x), x))
            True
            sage: diff(sec(x), x)
            sec(x)*tan(x)
            sage: latex(sec(x))
            \sec\left(x\right)
            sage: sec(x)._sympy_()                                                      # needs sympy
            sec(x)

        We can prevent evaluation using the ``hold`` parameter::

            sage: sec(pi/4, hold=True)                                                  # needs sage.symbolic
            sec(1/4*pi)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: a = sec(pi/4, hold=True); a.simplify()                                # needs sage.symbolic
            sqrt(2)

        TESTS:

        Test complex input::

            sage: sec(complex(1,1))     # rel tol 1e-15                                 # needs sage.rings.complex_double
            (0.49833703055518686+0.5910838417210451j)
        """
        GinacFunction.__init__(self, 'sec', latex_name=r"\sec",
                               conversions=dict(maxima='sec', mathematica='Sec',
                                 giac='sec', fricas='sec', sympy='sec'))

    def _eval_numpy_(self, x):
        """
        EXAMPLES::

            sage: import numpy                                                          # needs numpy
            sage: a = numpy.arange(2, 5)                                                # needs numpy
            sage: sec(a)                                                                # needs numpy
            array([-2.40299796, -1.01010867, -1.52988566])
        """
        return 1 / cos(x)


sec = Function_sec()


class Function_csc(GinacFunction):
    def __init__(self):
        r"""
        The cosecant function.

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: csc(pi/4)
            sqrt(2)
            sage: csc(x).subs(x==pi/4)
            sqrt(2)
            sage: csc(pi/7)
            csc(1/7*pi)
            sage: csc(x)
            csc(x)
            sage: RR(csc(pi/4))
            1.41421356237310
            sage: n(csc(pi/4), 100)
            1.4142135623730950488016887242
            sage: float(csc(pi/4))
            1.4142135623730951
            sage: csc(1/2)
            csc(1/2)
            sage: csc(0.5)
            2.08582964293349

            sage: # needs sage.symbolic
            sage: bool(diff(csc(x), x) == diff(1/sin(x), x))
            True
            sage: diff(csc(x), x)
            -cot(x)*csc(x)
            sage: latex(csc(x))
            \csc\left(x\right)
            sage: csc(x)._sympy_()                                                      # needs sympy
            csc(x)

        We can prevent evaluation using the ``hold`` parameter::

            sage: csc(pi/4, hold=True)                                                  # needs sage.symbolic
            csc(1/4*pi)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: a = csc(pi/4,hold=True); a.simplify()                                 # needs sage.symbolic
            sqrt(2)

        TESTS:

        Test complex input::

            sage: csc(complex(1,1))     # rel tol 1e-15                                 # needs sage.rings.complex_double
            (0.6215180171704284-0.30393100162842646j)
        """
        GinacFunction.__init__(self, 'csc', latex_name=r"\csc",
                               conversions=dict(maxima='csc', mathematica='Csc',
                                 giac='csc', fricas='csc', sympy='csc'))

    def _eval_numpy_(self, x):
        """
        EXAMPLES::

            sage: import numpy                                                          # needs numpy
            sage: a = numpy.arange(2, 5)                                                # needs numpy
            sage: csc(a)                                                                # needs numpy
            array([ 1.09975017,  7.0861674 , -1.32134871])
        """
        return 1 / sin(x)


csc = Function_csc()


###################################
# Inverse Trigonometric Functions #
###################################

class Function_arcsin(GinacFunction):
    def __init__(self):
        """
        The arcsine function.

        EXAMPLES::

            sage: arcsin(0.5)                                                           # needs sage.rings.real_mpfr
            0.523598775598299
            sage: arcsin(1/2)                                                           # needs sage.symbolic
            1/6*pi
            sage: arcsin(1 + 1.0*I)                                                     # needs sage.symbolic
            0.666239432492515 + 1.06127506190504*I

        We can delay evaluation using the ``hold`` parameter::

            sage: arcsin(0, hold=True)                                                  # needs sage.symbolic
            arcsin(0)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: a = arcsin(0, hold=True); a.simplify()                                # needs sage.symbolic
            0

        ``conjugate(arcsin(x))==arcsin(conjugate(x))``, unless on the branch
        cuts which run along the real axis outside the interval [-1, +1].::

            sage: # needs sage.symbolic
            sage: conjugate(arcsin(x))
            conjugate(arcsin(x))
            sage: var('y', domain='positive')
            y
            sage: conjugate(arcsin(y))
            conjugate(arcsin(y))
            sage: conjugate(arcsin(y+I))
            conjugate(arcsin(y + I))
            sage: conjugate(arcsin(1/16))
            arcsin(1/16)
            sage: conjugate(arcsin(2))
            conjugate(arcsin(2))
            sage: conjugate(arcsin(-2))
            -conjugate(arcsin(2))

        TESTS::

            sage: arcsin(x)._sympy_()                                                   # needs sympy sage.symbolic
            asin(x)
            sage: arcsin(x).operator()                                                  # needs sage.symbolic
            arcsin
            sage: asin(complex(1,1))                                                    # needs sage.rings.complex_double
            (0.6662394324925152+1.0612750619050357j)
            sage: asin(SR(2.1))                                                         # needs sage.symbolic
            1.57079632679490 - 1.37285914424258*I
        """
        GinacFunction.__init__(self, 'arcsin', latex_name=r"\arcsin",
                conversions=dict(maxima='asin', sympy='asin',
                                 mathematica='ArcSin',
                                 fricas='asin', giac='asin'))


arcsin = asin = Function_arcsin()


class Function_arccos(GinacFunction):
    def __init__(self):
        """
        The arccosine function.

        EXAMPLES::

            sage: arccos(0.5)                                                           # needs sage.rings.real_mpfr
            1.04719755119660
            sage: arccos(1/2)                                                           # needs sage.symbolic
            1/3*pi
            sage: arccos(1 + 1.0*I)                                                     # needs sage.symbolic
            0.904556894302381 - 1.06127506190504*I
            sage: arccos(3/4).n(100)                                                    # needs sage.symbolic
            0.72273424781341561117837735264

        We can delay evaluation using the ``hold`` parameter::

            sage: arccos(0, hold=True)                                                  # needs sage.symbolic
            arccos(0)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: a = arccos(0, hold=True); a.simplify()                                # needs sage.symbolic
            1/2*pi

        ``conjugate(arccos(x))==arccos(conjugate(x))``, unless on the branch
        cuts, which run along the real axis outside the interval [-1, +1].::

            sage: # needs sage.symbolic
            sage: conjugate(arccos(x))
            conjugate(arccos(x))
            sage: var('y', domain='positive')
            y
            sage: conjugate(arccos(y))
            conjugate(arccos(y))
            sage: conjugate(arccos(y+I))
            conjugate(arccos(y + I))
            sage: conjugate(arccos(1/16))
            arccos(1/16)
            sage: conjugate(arccos(2))
            conjugate(arccos(2))
            sage: conjugate(arccos(-2))
            pi - conjugate(arccos(2))

        TESTS::

            sage: arccos(x)._sympy_()                                                   # needs sympy sage.symbolic
            acos(x)
            sage: arccos(x).operator()                                                  # needs sage.symbolic
            arccos
            sage: acos(complex(1,1))                                                    # needs sage.rings.complex_double
            (0.9045568943023814-1.0612750619050357j)
            sage: acos(SR(2.1))                                                         # needs sage.symbolic
            1.37285914424258*I
        """
        GinacFunction.__init__(self, 'arccos', latex_name=r"\arccos",
                conversions=dict(maxima='acos', sympy='acos',
                                 mathematica='ArcCos',
                                 fricas='acos', giac='acos'))


arccos = acos = Function_arccos()


class Function_arctan(GinacFunction):
    def __init__(self):
        """
        The arctangent function.

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: arctan(1/2)
            arctan(1/2)
            sage: RDF(arctan(1/2))  # rel tol 1e-15
            0.46364760900080615
            sage: arctan(1 + I)
            arctan(I + 1)
            sage: arctan(1/2).n(100)
            0.46364760900080611621425623146

        We can delay evaluation using the ``hold`` parameter::

            sage: arctan(0, hold=True)                                                  # needs sage.symbolic
            arctan(0)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: a = arctan(0, hold=True); a.simplify()                                # needs sage.symbolic
            0

        ``conjugate(arctan(x))==arctan(conjugate(x))``, unless on the branch
        cuts which run along the imaginary axis outside the interval [-I, +I].::

            sage: # needs sage.symbolic
            sage: conjugate(arctan(x))
            conjugate(arctan(x))
            sage: var('y', domain='positive')
            y
            sage: conjugate(arctan(y))
            arctan(y)
            sage: conjugate(arctan(y+I))
            conjugate(arctan(y + I))
            sage: conjugate(arctan(1/16))
            arctan(1/16)
            sage: conjugate(arctan(-2*I))
            conjugate(arctan(-2*I))
            sage: conjugate(arctan(2*I))
            conjugate(arctan(2*I))
            sage: conjugate(arctan(I/2))
            arctan(-1/2*I)

        TESTS::

            sage: arctan(x)._sympy_()                                                   # needs sympy sage.symbolic
            atan(x)
            sage: arctan(x).operator()                                                  # needs sage.symbolic
            arctan
            sage: atan(complex(1,1))                                                    # needs sage.rings.complex_double
            (1.0172219678978514+0.4023594781085251j)

        Check that :issue:`19918` is fixed::

            sage: arctan(-x).subs(x=oo)                                                 # needs sage.symbolic
            -1/2*pi
            sage: arctan(-x).subs(x=-oo)                                                # needs sage.symbolic
            1/2*pi
        """
        GinacFunction.__init__(self, 'arctan', latex_name=r"\arctan",
                conversions=dict(maxima='atan', sympy='atan',
                                 mathematica='ArcTan',
                                 fricas='atan', giac='atan'))


arctan = atan = Function_arctan()


class Function_arccot(GinacFunction):
    def __init__(self):
        """
        The arccotangent function.

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: arccot(1/2)
            arccot(1/2)
            sage: RDF(arccot(1/2))  # abs tol 2e-16
            1.1071487177940906
            sage: arccot(1 + I)
            arccot(I + 1)
            sage: arccot(1/2).n(100)
            1.1071487177940905030170654602
            sage: float(arccot(1/2))  # abs tol 2e-16
            1.1071487177940906
            sage: bool(diff(acot(x), x) == -diff(atan(x), x))
            True
            sage: diff(acot(x), x)
            -1/(x^2 + 1)

        We can delay evaluation using the ``hold`` parameter::

            sage: arccot(1, hold=True)                                                  # needs sage.symbolic
            arccot(1)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: a = arccot(1, hold=True); a.simplify()                                # needs sage.symbolic
            1/4*pi

        TESTS:

        Test complex input::

            sage: arccot(x)._sympy_()                                                   # needs sympy sage.symbolic
            acot(x)
            sage: arccot(complex(1,1))  # rel tol 1e-15                                 # needs sage.rings.complex_double
            (0.5535743588970452-0.4023594781085251j)
            sage: arccot(1.+I)                                                          # needs sage.symbolic
            0.553574358897045 - 0.402359478108525*I
        """
        GinacFunction.__init__(self, 'arccot', latex_name=r"\operatorname{arccot}",
                conversions=dict(maxima='acot', sympy='acot',
                                 mathematica='ArcCot',
                                 fricas='acot', giac='acot'))

    def _eval_numpy_(self, x):
        """
        EXAMPLES::

            sage: import numpy                                                          # needs numpy
            sage: a = numpy.arange(2, 5)                                                # needs numpy
            sage: arccot(a)                                                             # needs numpy
            array([0.46364761, 0.32175055, 0.24497866])
        """
        return math.pi / 2 - arctan(x)


arccot = acot = Function_arccot()


class Function_arccsc(GinacFunction):
    def __init__(self):
        """
        The arccosecant function.

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: arccsc(2)
            arccsc(2)
            sage: RDF(arccsc(2))  # rel tol 1e-15
            0.5235987755982988
            sage: arccsc(2).n(100)
            0.52359877559829887307710723055
            sage: float(arccsc(2))
            0.52359877559829...
            sage: arccsc(1 + I)
            arccsc(I + 1)
            sage: diff(acsc(x), x)
            -1/(sqrt(x^2 - 1)*x)
            sage: arccsc(x)._sympy_()                                                   # needs sympy
            acsc(x)

        We can delay evaluation using the ``hold`` parameter::

            sage: arccsc(1, hold=True)                                                  # needs sage.symbolic
            arccsc(1)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: a = arccsc(1, hold=True); a.simplify()                                # needs sage.symbolic
            1/2*pi

        TESTS:

        Test complex input::

            sage: arccsc(complex(1,1))  # rel tol 1e-15                                 # needs sage.rings.complex_double
            (0.45227844715119064-0.5306375309525178j)
        """
        GinacFunction.__init__(self, 'arccsc', latex_name=r"\operatorname{arccsc}",
                               conversions=dict(maxima='acsc', sympy='acsc',
                                                mathematica='ArcCsc',
                                                fricas='acsc', giac='acsc'))

    def _eval_numpy_(self, x):
        """
        EXAMPLES::

            sage: import numpy                                                          # needs numpy
            sage: a = numpy.arange(2, 5)                                                # needs numpy
            sage: arccsc(a)                                                             # needs numpy
            array([0.52359878, 0.33983691, 0.25268026])
        """
        return arcsin(1.0 / x)


arccsc = acsc = Function_arccsc()


class Function_arcsec(GinacFunction):
    def __init__(self):
        """
        The arcsecant function.

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: arcsec(2)
            arcsec(2)
            sage: arcsec(2.0)
            1.04719755119660
            sage: arcsec(2).n(100)
            1.0471975511965977461542144611
            sage: arcsec(1/2).n(100)
            1.3169578969248167086250463473*I
            sage: RDF(arcsec(2))  # abs tol 1e-15
            1.0471975511965976
            sage: arcsec(1 + I)
            arcsec(I + 1)
            sage: diff(asec(x), x)
            1/(sqrt(x^2 - 1)*x)
            sage: arcsec(x)._sympy_()                                                   # needs sympy
            asec(x)

        We can delay evaluation using the ``hold`` parameter::

            sage: arcsec(1, hold=True)                                                  # needs sage.symbolic
            arcsec(1)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: a = arcsec(1, hold=True); a.simplify()                                # needs sage.symbolic
            0

        TESTS:

        Test complex input::

            sage: arcsec(complex(1,1))  # rel tol 1e-15                                 # needs sage.rings.complex_double
            (1.118517879643706+0.5306375309525178j)
        """
        GinacFunction.__init__(self, 'arcsec', latex_name=r"\operatorname{arcsec}",
                               conversions=dict(maxima='asec', sympy='asec',
                                                mathematica='ArcSec',
                                                fricas='asec', giac='asec'))

    def _eval_numpy_(self, x):
        """
        EXAMPLES::

            sage: import numpy                                                          # needs numpy
            sage: a = numpy.arange(2, 5)                                                # needs numpy
            sage: arcsec(a)                                                             # needs numpy
            array([1.04719755, 1.23095942, 1.31811607])
        """
        return arccos(1.0 / x)


arcsec = asec = Function_arcsec()


class Function_arctan2(GinacFunction):
    def __init__(self):
        r"""
        The modified arctangent function.

        Returns the arc tangent (measured in radians) of `y/x`, where
        unlike ``arctan(y/x)``, the signs of both ``x`` and ``y`` are
        considered.  In particular, this function measures the angle
        of a ray through the origin and `(x,y)`, with the positive
        `x`-axis the zero mark, and with output angle `\theta`
        being between `-\pi<\theta<=\pi`.

        Hence, ``arctan2(y,x) = arctan(y/x)`` only for `x>0`.  One
        may consider the usual arctan to measure angles of lines
        through the origin, while the modified function measures
        rays through the origin.

        Note that the `y`-coordinate is by convention the first input.

        EXAMPLES:

        Note the difference between the two functions::

            sage: arctan2(1, -1)                                                        # needs sage.symbolic
            3/4*pi
            sage: arctan(1/-1)                                                          # needs sage.symbolic
            -1/4*pi

        This is consistent with Python and Maxima::

            sage: maxima.atan2(1, -1)                                                   # needs sage.symbolic
            (3*%pi)/4
            sage: math.atan2(1, -1)
            2.356194490192345

        More examples::

            sage: arctan2(1, 0)                                                         # needs sage.symbolic
            1/2*pi
            sage: arctan2(2, 3)                                                         # needs sage.symbolic
            arctan(2/3)
            sage: arctan2(-1, -1)                                                       # needs sage.symbolic
            -3/4*pi

        Of course we can approximate as well::

            sage: arctan2(-1/2, 1).n(100)                                               # needs sage.symbolic
            -0.46364760900080611621425623146
            sage: arctan2(2, 3).n(100)                                                  # needs sage.symbolic
            0.58800260354756755124561108063

        We can delay evaluation using the ``hold`` parameter::

            sage: arctan2(-1/2, 1, hold=True)                                           # needs sage.symbolic
            arctan2(-1/2, 1)

        To then evaluate again, we currently must use Maxima via
        :meth:`sage.symbolic.expression.Expression.simplify`::

            sage: arctan2(-1/2, 1, hold=True).simplify()                                # needs sage.symbolic
            -arctan(1/2)

        The function also works with numpy arrays as input::

            sage: # needs numpy
            sage: import numpy
            sage: a = numpy.linspace(1, 3, 3)
            sage: b = numpy.linspace(3, 6, 3)
            sage: atan2(a, b)
            array([0.32175055, 0.41822433, 0.46364761])

            sage: atan2(1,a)                                                            # needs numpy
            array([0.78539816, 0.46364761, 0.32175055])

            sage: atan2(a, 1)                                                           # needs numpy
            array([0.78539816, 1.10714872, 1.24904577])

        TESTS::

            sage: x,y = var('x,y')                                                      # needs sage.symbolic
            sage: arctan2(y, x).operator()                                              # needs sage.symbolic
            arctan2

        Check if :issue:`8565` is fixed::

            sage: atan2(-pi, 0)                                                         # needs sage.symbolic
            -1/2*pi

        Check if :issue:`8564` is fixed::

            sage: arctan2(x,x)._sympy_()                                                # needs sympy sage.symbolic
            atan2(x, x)

        Check if numerical evaluation works :issue:`9913`::

            sage: arctan2(0, -log(2)).n()                                               # needs sage.symbolic
            3.14159265358979

        Check that atan2(0,0) returns NaN :issue:`21614`::

            sage: # needs sage.symbolic
            sage: atan2(0, 0)
            NaN
            sage: atan2(0, 0).n()
            NaN
            sage: atan2(0, 0, hold=True)
            arctan2(0, 0)
            sage: atan2(0, 0, hold=True).n()
            Traceback (most recent call last):
            ...
            RuntimeError: atan2(): division by zero

        Check if :issue:`10062` is fixed, this was caused by
        ``(I*I).is_positive()`` returning ``True``::

            sage: arctan2(0, I*I)                                                       # needs sage.symbolic
            pi
        """
        GinacFunction.__init__(self, 'arctan2', nargs=2, latex_name=r"\arctan",
                conversions=dict(maxima='atan2', sympy='atan2', giac='atan2'))


arctan2 = atan2 = Function_arctan2()



# sage_setup: distribution = sagemath-categories
from sage.rings.finite_rings.finite_field_constructor import FiniteField

GF = FiniteField

from sage.rings.finite_rings.conway_polynomials import conway_polynomial, exists_conway_polynomial

# Finite residue fields
from sage.rings.finite_rings.residue_field import ResidueField



# sage_setup: distribution = sagemath-categories
"""
Sparse bitset.

This is a regular bitset to which we will add additional structure.

In particular some representation of which limbs even contain data.
"""
# ****************************************************************************
#       Copyright (C) 2020 Jonathan Kliem <jonathan.kliem@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.libs.gmp.types cimport *

cdef struct sparse_bitset_s:
    # The size of a bitset B counts the maximum number of bits that B can
    # hold. This size is independent of how many elements of B are toggled to
    # 1. For example, say B is the bitset 1001. Then B has size 4, with the
    # first and fourth elements toggled to 1, reading from left to right.
    # We can also think of the size of a bitset as its capacity.
    mp_bitcnt_t size

    # A limb is that part of a bitset that can fit into an mp_limb_t
    # (typically, 32 bits on a 32-bit machine and 64 bits on a 64-bit
    # machine). This counts the number of limbs to represent a bitset.
    # If a bitset has size <= n, then the whole bitset fits into a limb
    # and we only require one limb to represent the bitset. However, if
    # the bitset has size > n, we require more than one limb to
    # represent the bitset. For example, if a limb is 64 bits in length
    # and the bitset has size 96 bits, then we require at most two limbs
    # to represent the bitset.
    #
    # NOTE: some code assumes that mp_limb_t is an unsigned long
    # (this assumption is always true in practice).
    mp_size_t limbs

    # The individual bits of a bitset.
    # NOTE: ``sparse_bitset_t`` is assumed to be allocated over-aligned.
    mp_limb_t* bits

    # Pointer to the memory of ``bits``.
    void* mem

    # Storing the nonzero positions can safe time, when performing
    # multiple comparisons.
    # E.g. one can set them while computing the intersection
    # and then use those to ``bitset_issubset`` many times in a row.

    # Any modification, will invalidate the already computed positions.
    # It is stored, whether the nonzero chunks are correctly initialized
    # or not. Computations will work correctly either way.
    bint non_zero_chunks_are_initialized
    mp_bitcnt_t* non_zero_chunks
    mp_bitcnt_t n_non_zero_chunks

ctypedef sparse_bitset_s sparse_bitset_t[1]



# sage_setup: distribution = sagemath-categories
from sage.misc.lazy_import import lazy_import

lazy_import('sage.schemes.product_projective.space', 'ProductProjectiveSpaces')
lazy_import('sage.schemes.product_projective.space', 'is_ProductProjectiveSpaces')
del lazy_import



# sage_setup: distribution = sagemath-categories
r"""
Sets With a Grading
"""
# ****************************************************************************
#  Copyright (C) 2010-2012 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.categories.category_types import Category
from sage.categories.enumerated_sets import EnumeratedSets
from sage.categories.sets_cat import Sets
from sage.misc.abstract_method import abstract_method
from sage.misc.cachefunc import cached_method


class SetsWithGrading(Category):
    r"""
    The category of sets with a grading.

    A *set with a grading* is a set `S` equipped with a
    grading by some other set `I` (by default the set `\NN` of the
    nonnegative integers):

    .. MATH::

         S = \biguplus_{i\in I} S_i

    where the *graded components* `S_i` are (usually finite)
    sets. The *grading* function maps each element `s` of
    `S` to its *grade* `i`, so that `s\in S_i`.

    From implementation point of view, if the graded set is enumerated then
    each graded component should be enumerated (there is a check in the method
    :meth:`~SetsWithGrading.ParentMethods._test_graded_components`). The
    contrary needs not be true.

    To implement this category, a parent must either implement
    :meth:`~SetsWithGrading.ParentMethods.graded_component()` or
    :meth:`~SetsWithGrading.ParentMethods.subset()`. If only
    :meth:`~SetsWithGrading.ParentMethods.subset()` is implemented, the first
    argument must be the grading for compatibility with
    :meth:`~SetsWithGrading.ParentMethods.graded_component()`. Additionally
    either the parent must implement
    :meth:`~SetsWithGrading.ParentMethods.grading()` or its elements must
    implement a method ``grade()``. See the example
    :class:`sage.categories.examples.sets_with_grading.NonNegativeIntegers`.

    Finally, if the graded set is enumerated (see
    :class:`~sage.categories.enumerated_sets.EnumeratedSets`) then each graded
    component should be enumerated. The contrary needs not be true.

    EXAMPLES:

    A typical example of a set with a grading is the set of nonnegative
    integers graded by themselves::

        sage: N = SetsWithGrading().example(); N
        Non negative integers
        sage: N.category()
        Category of facade infinite sets with grading
        sage: N.grading_set()
        Non negative integers

    The *grading function* is given by ``N.grading``::

        sage: N.grading(4)
        4

    The graded component `N_i` is the set with one element `i`::

        sage: N.graded_component(grade=5)
        {5}
        sage: N.graded_component(grade=42)
        {42}

    Here are some information about this category::

        sage: SetsWithGrading()
        Category of sets with grading
        sage: SetsWithGrading().super_categories()
        [Category of sets]
        sage: SetsWithGrading().all_super_categories()
        [Category of sets with grading,
         Category of sets,
         Category of sets with partial maps,
         Category of objects]

    .. TODO::

        - This should be moved to ``Sets().WithGrading()``.
        - Should the grading set be a parameter for this category?
        - Does the enumeration need to be compatible with the grading? Be
          careful that the fact that graded components are allowed to be finite
          or infinite make the answer complicated.

    TESTS::

        sage: C = SetsWithGrading()
        sage: TestSuite(C).run()
    """

    @cached_method
    def super_categories(self):
        """
        EXAMPLES::

            sage: SetsWithGrading().super_categories()
            [Category of sets]
        """
        return [Sets()]

    class ParentMethods:

        def _test_graded_components(self, **options):
            r"""
            Test that some graded components of ``self`` are parent with
            initialized category and that the parent has a properly implemented
            ``grading()`` method.

            EXAMPLES::

                sage: SetsWithGrading().example()._test_graded_components()
            """
            tester = self._tester(**options)
            for grade in self.grading_set().some_elements():
                G = self.graded_component(grade)
                if self in EnumeratedSets():
                    tester.assertIn(G, EnumeratedSets())
                else:
                    tester.assertIn(G, Sets())
                for elt in G.some_elements():
                    tester.assertEqual(self.grading(elt), grade)

        def grading_set(self):
            """
            Return the set ``self`` is graded by. By default, this is
            the set of nonnegative integers.

            EXAMPLES::

                sage: SetsWithGrading().example().grading_set()
                Non negative integers
            """
            from sage.sets.non_negative_integers import NonNegativeIntegers
            return NonNegativeIntegers()

        # TODO:
        #  - Should this method be in EnumeratedSets? With a default implementation
        #    a la ``filter``?
        #  - Do we want to enforce implementing subset rather than graded_component?
        @abstract_method(optional=True)
        def subset(self, *args, **options):
            """
            Return the subset of ``self`` described by the given parameters.

            .. SEEALSO::

                -:meth:`graded_component()`

            EXAMPLES::

                sage: W = WeightedIntegerVectors([3,2,1]); W
                Integer vectors weighted by [3, 2, 1]
                sage: W.subset(4)
                Integer vectors of 4 weighted by [3, 2, 1]
            """

        def graded_component(self, grade):
            """
            Return the graded component of ``self`` with grade ``grade``.

            The default implementation just calls the method :meth:`subset()`
            with the first argument ``grade``.

            EXAMPLES::

                sage: N = SetsWithGrading().example(); N
                Non negative integers
                sage: N.graded_component(3)
                {3}
            """
            return self.subset(grade)

        def grading(self, elt):
            """
            Return the grading of the element ``elt`` of ``self``.

            This default implementation calls ``elt.grade()``.

            EXAMPLES::

                sage: N = SetsWithGrading().example(); N
                Non negative integers
                sage: N.grading(4)
                4
            """
            return elt.grade()

        def generating_series(self):
            """
            Default implementation for generating series.

            OUTPUT: a series, indexed by the grading set

            EXAMPLES::

                sage: N = SetsWithGrading().example(); N
                Non negative integers
                sage: N.generating_series()
                1/(-z + 1)

                sage: Permutations().generating_series()                                # needs sage.combinat
                1 + z + 2*z^2 + 6*z^3 + 24*z^4 + 120*z^5 + 720*z^6 + O(z^7)

             .. TODO::

                 - Very likely, this should always return a lazy power series.
            """
            from sage.rings.integer_ring import ZZ
            from sage.rings.lazy_series_ring import LazyPowerSeriesRing
            from sage.sets.non_negative_integers import NonNegativeIntegers
            if isinstance(self.grading_set(), NonNegativeIntegers):
                R = LazyPowerSeriesRing(ZZ, names='z')
                return R(lambda n: self.graded_component(n).cardinality())
            raise NotImplementedError

        # TODO:
        #   * asymptotic behavior: we need an object for asymptotic behavior and
        #   a default name for the method that should be here. Such method will
        #   have two goals (and perhaps need two implementations): give a
        #   theorem on asymptotic and be a tool to determine a strategy for
        #   algorithms.
